from __future__ import annotations

import asyncio
import contextlib
import datetime
import operator
import os
import pathlib
import random
from collections.abc import AsyncIterator, Awaitable, Callable
from types import MethodType
from typing import Any

import aiohttp
import aiohttp_client_cache
import aiopath
import asyncstdlib
import discord
import discord.ext.commands
import ujson
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from asyncspotify import Client as SpotifyClient
from asyncspotify import ClientCredentialsFlow
from asyncstdlib import itertools
from discord.abc import Messageable
from discord.ext.commands import Context
from discord.types.embed import EmbedType
from packaging.version import Version

from pylav import VERSION

# noinspection PyProtectedMember
from pylav._internals.functions import add_property
from pylav.constants import MAX_RECURSION_DEPTH
from pylav.constants.config import (
    CONFIG_DIR,
    EXTERNAL_UNMANAGED_HOST,
    EXTERNAL_UNMANAGED_PASSWORD,
    EXTERNAL_UNMANAGED_PORT,
    EXTERNAL_UNMANAGED_SSL,
    JAVA_EXECUTABLE,
    MANAGED_NODE_APPLE_MUSIC_API_KEY,
    MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE,
    MANAGED_NODE_DEEZER_KEY,
    MANAGED_NODE_SPOTIFY_CLIENT_ID,
    MANAGED_NODE_SPOTIFY_CLIENT_SECRET,
    MANAGED_NODE_SPOTIFY_COUNTRY_CODE,
    MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN,
    PREFER_PARTIAL_TRACKS,
    REDIS_FULL_ADDRESS_RESPONSE_CACHE,
    TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
)
from pylav.constants.playlists import BUNDLED_DEEZER_PLAYLIST_IDS, BUNDLED_PLAYLIST_IDS, BUNDLED_SPOTIFY_PLAYLIST_IDS
from pylav.core.bot_overrides import get_context, process_commands
from pylav.core.context import PyLavContext
from pylav.events.base import PyLavEvent
from pylav.events.manager import DispatchManager
from pylav.exceptions.client import AnotherClientAlreadyRegisteredException, PyLavInvalidArgumentsException
from pylav.exceptions.node import NoNodeAvailableException, NoNodeWithRequestFunctionalityAvailableException
from pylav.exceptions.request import HTTPException
from pylav.extension.bundled_node import LAVALINK_DOWNLOAD_DIR
from pylav.extension.bundled_node.manager import LocalNodeManager
from pylav.extension.flowery.base import FloweryAPI
from pylav.extension.m3u import M3UParser
from pylav.extension.radio import RadioBrowser
from pylav.helpers.misc import MISSING
from pylav.helpers.singleton import SingletonCallable, SingletonClass
from pylav.helpers.time import get_now_utc, get_tz_utc
from pylav.logging import getLogger
from pylav.nodes.api.responses.rest_api import LoadTrackResponses
from pylav.nodes.api.responses.route_planner import Status as RoutePlannerStatus
from pylav.nodes.api.responses.track import Track as TrackObject
from pylav.nodes.manager import NodeManager
from pylav.nodes.node import Node
from pylav.players.manager import PlayerController
from pylav.players.player import Player
from pylav.players.query.obj import Query
from pylav.players.tracks.decoder import async_decoder
from pylav.players.tracks.obj import Track
from pylav.storage.controllers.config import ConfigController
from pylav.storage.controllers.equalizers import EqualizerController
from pylav.storage.controllers.migrator import MigrationController
from pylav.storage.controllers.nodes import NodeController
from pylav.storage.controllers.players.config import PlayerConfigController
from pylav.storage.controllers.players.states import PlayerStateController
from pylav.storage.controllers.playlists import PlaylistController
from pylav.storage.controllers.queries import QueryController
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.misc import DATABASE_ENGINE, IS_POSTGRES
from pylav.storage.models.config import Config
from pylav.storage.models.equilizer import Equalizer
from pylav.storage.models.node.real import Node as RealNode
from pylav.storage.models.player.state import PlayerState
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE, DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE
from pylav.utils.aiohttp_postgres_cache import PostgresCacheBackend
from pylav.utils.vendor.redbot import AsyncIter

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLav", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


LOGGER = getLogger("PyLav.Client")


class Client(metaclass=SingletonClass):
    """
    Represents a Lavalink client used to manage nodes and connections.

    .. _event loop: https://docs.python.org/3/library/asyncio-eventloop.html

    Parameters
    ----------
    bot : :class:`DISCORD_BOT_TYPE`
        The bot instance.
    cog: :class:`DISCORD_COG_TYPE`
        The cog to register the client to.
    player: Optional[:class:`Player`]
        The class that should be used for the player. Defaults to ``Player``.
        Do not change this unless you know what you are doing!
    connect_back: Optional[:class:`bool`]
        A boolean that determines if a player will connect back to the
        node it was originally connected to. This is not recommended doing since
        the player will most likely be performing better in the new node. Defaults to `False`.

        Warning
        -------
        If this option is enabled and the player's node is changed through `Player.change_node` after
        the player was moved via the fail-over mechanism, the player will still move back to the original
        node when it becomes available. This behaviour can be avoided in custom player implementations by
        setting `self._original_node` to `None` in the `change_node` function.
    config_folder: Optional[:class:`pathlib.Path`]
        The path to the config folder. Defaults to ``CONFIG_DIR``.
    """

    _local_node_manager: LocalNodeManager
    _asyncio_lock = asyncio.Lock()
    _config: Config
    __cogs_registered = set()
    __old_process_command_method: Callable = None
    __old_get_context: Callable = None
    _initiated = False

    def __init__(
        self,
        bot: DISCORD_BOT_TYPE,
        cog: DISCORD_COG_TYPE,
        player: type[Player] = Player,  # type: ignore
        connect_back: bool = False,
        config_folder: aiopath.AsyncPath | pathlib.Path = CONFIG_DIR,
    ):
        try:
            setattr(bot, "_pylav_client", self)
            # noinspection PyProtectedMember
            add_property(bot, "lavalink", lambda b: b._pylav_client)
            # noinspection PyProtectedMember
            add_property(bot, "pylav", lambda b: b._pylav_client)
            if self.__old_process_command_method is None:
                self.__old_process_command_method = bot.process_commands
            if self.__old_get_context is None:
                self.__old_get_context = bot.get_context
            self._disable_translations = False
            self._context_translator: Callable[[DISCORD_BOT_TYPE, discord.Guild], Awaitable[None]] | None = None
            self.ready = asyncio.Event()
            bot.process_commands = MethodType(process_commands, bot)
            bot.get_context = MethodType(get_context, bot)
            self.__cogs_registered.add(cog.__cog_name__)
            config_folder = pathlib.Path(config_folder)
            (config_folder / ".data").mkdir(exist_ok=True, parents=True)
            self._config_folder = aiopath.AsyncPath(config_folder)
            self._bot = bot
            self._user_id = str(bot.user.id)
            if REDIS_FULL_ADDRESS_RESPONSE_CACHE:
                self._aiohttp_client_cache = aiohttp_client_cache.RedisBackend(
                    address=REDIS_FULL_ADDRESS_RESPONSE_CACHE,
                    cache_control=True,
                    allowed_codes=(200,),
                    allowed_methods=("GET",),
                    include_headers=True,
                    expire_after=datetime.timedelta(days=1),
                    timeout=2.5,
                    ignored_params=["auth_token", "timestamp"],
                )
            else:
                self._aiohttp_client_cache = PostgresCacheBackend(
                    cache_control=True,
                    allowed_codes=(200,),
                    allowed_methods=("GET",),
                    include_headers=True,
                    ignored_params=["auth_token", "timestamp"],
                    expire_after=datetime.timedelta(days=1),
                    timeout=2.5,
                )
            self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps)
            self._cached_session = aiohttp_client_cache.CachedSession(
                timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps, cache=self._aiohttp_client_cache
            )
            # Attach the Client to the necessary objects
            CachedModel.attach_client(self)
            Query.attach_client(self)
            Track.attach_client(self)
            PlayerState.attach_client(self)
            Equalizer.attach_client(self)

            self._node_manager = NodeManager(
                self,
                external_host=EXTERNAL_UNMANAGED_HOST,
                external_port=EXTERNAL_UNMANAGED_PORT,
                external_password=EXTERNAL_UNMANAGED_PASSWORD,
                external_ssl=EXTERNAL_UNMANAGED_SSL,
            )
            self._player_manager = PlayerController(self, player)
            self._lib_config_manager = ConfigController(self)
            self._node_config_manager = NodeController(self)
            self._playlist_config_manager = PlaylistController(self)
            self._query_cache_manager = QueryController(self)
            self._update_schema_manager = MigrationController(self)
            self._dispatch_manager = DispatchManager(self)
            self._player_state_db_manager = PlayerStateController(self)
            self._player_config_manager = PlayerConfigController(self)
            self._equalizer_config_manager = EqualizerController(self)
            self._flowery_api = FloweryAPI(self)
            self._radio_manager = RadioBrowser(self)
            self._m3u8parser = M3UParser(self)
            self._connect_back = connect_back
            self._warned_about_no_search_nodes = False
            self._spotify_client_id = None
            self._spotify_client_secret = None
            self._spotify_auth = None
            self._shutting_down = False
            self._scheduler = AsyncIOScheduler()
            self._scheduler.configure(timezone=get_tz_utc())
            self._wait_for_playlists = asyncio.Event()
            self._wait_for_playlists.set()
        except Exception:
            LOGGER.exception("Failed to initialize Lavalink")
            raise

    async def set_context_locale(self, guild: discord.Guild | None) -> None:
        if self._disable_translations:
            return
        try:
            if self._context_translator is None:
                from redbot.core.i18n import set_contextual_locales_from_guild

                self._context_translator = set_contextual_locales_from_guild
            await self._context_translator(self.bot, guild)
        except Exception:  # noqa
            self._disable_translations = True

    async def on_pylav_red_api_tokens_update(self, service_name: str, api_tokens: dict[str, str]) -> None:
        match service_name:
            case "spotify" if "client_id" in api_tokens and "client_secret" in api_tokens:
                await self.update_spotify_tokens(**api_tokens)
            case "apple_music" if "token" in api_tokens and "country_code" in api_tokens:
                await self.update_applemusic_tokens(**api_tokens)
            case "deezer" if "master_token" in api_tokens:
                await self.update_deezer_tokens(**api_tokens)
            case "yandexmusic" if "token" in api_tokens:
                await self.update_yandex_tokens(**api_tokens)
            case "google" if ("email" in api_tokens and "password" in api_tokens):
                await self.update_google_account(**api_tokens)

    async def on_pylav_shard_resumed(self, shard_id: int) -> None:
        if self._shutting_down or not self.initialized:
            return
        LOGGER.debug("Shard %s resumed, checking for affected players", shard_id)
        players = filter(lambda p: p.guild.shard_id == shard_id, self.player_manager.players.values())
        for player in players:
            await self.set_context_locale(player.guild)
            await player.reconnect()

    async def on_pylav_shard_ready(self, shard_id: int) -> None:
        if self._shutting_down or not self.initialized:
            return
        LOGGER.debug("Shard %s ready, checking for affected players", shard_id)
        players = filter(lambda p: p.guild.shard_id == shard_id, self.player_manager.players.values())
        for player in players:
            await self.set_context_locale(player.guild)
            await player.reconnect()

    async def on_pylav_resumed(self) -> None:
        if self._shutting_down or not self.initialized:
            return
        LOGGER.debug("Resumed, checking for affected players")

        for player in self.player_manager.players.values():
            await self.set_context_locale(player.guild)
            await player.reconnect()

    async def on_pylav_ready(self) -> None:
        if self._shutting_down or not self.initialized:
            return
        LOGGER.debug("Ready, checking for affected players")
        for player in self.player_manager.players.values():
            await self.set_context_locale(player.guild)
            await player.reconnect()

    async def wait_until_ready(self, timeout: float | None = None) -> None:
        await asyncio.wait_for(self.ready.wait(), timeout=timeout)

    @property
    def initialized(self) -> bool:
        """Returns whether the client has been initialized"""
        return self._initiated

    @property
    def scheduler(self) -> AsyncIOScheduler:
        """Returns the scheduler"""
        return self._scheduler

    @property
    def radio_browser(self) -> RadioBrowser:
        """Returns the radio browser instance"""
        return self._radio_manager

    @property
    def dispatch_manager(self) -> DispatchManager:
        """Returns the dispatch manager"""
        return self._dispatch_manager

    @property
    def player_config_manager(self) -> PlayerConfigController:
        """Returns the player config manager"""
        return self._player_config_manager

    @property
    def spotify_client(self) -> SpotifyClient | None:
        """Returns the spotify client"""
        return SpotifyClient(self._spotify_auth) if self._spotify_auth else None

    @property
    def is_shutting_down(self) -> bool:
        """Returns whether the client is shutting down"""
        return self._shutting_down

    @property
    def node_db_manager(self) -> NodeController:
        """Returns the sql node config manager"""
        return self._node_config_manager

    @property
    def player_state_db_manager(self) -> PlayerStateController:
        return self._player_state_db_manager

    @property
    def playlist_db_manager(self) -> PlaylistController:
        """Returns the sql playlist config manager"""
        return self._playlist_config_manager

    @property
    def equalizer_db_manager(self) -> EqualizerController:
        """Returns the sql equalizer config manager"""
        return self._equalizer_config_manager

    @property
    def flowery_api(self) -> FloweryAPI:
        """Returns the flowery api"""
        return self._flowery_api

    @property
    def lib_db_manager(self) -> ConfigController:
        """Returns the sql lib config manager"""
        return self._lib_config_manager

    @property
    def query_cache_manager(self) -> QueryController:
        """Returns the query cache manager"""
        return self._query_cache_manager

    @property
    def managed_node_controller(self) -> LocalNodeManager:
        return self._local_node_manager

    @property
    def node_manager(self) -> NodeManager:
        return self._node_manager

    @property
    def player_manager(self) -> PlayerController:
        return self._player_manager

    @property
    def config_folder(self) -> aiopath.AsyncPath:
        return self._config_folder

    @property
    def bot(self) -> DISCORD_BOT_TYPE:
        return self._bot

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    @property
    def cached_session(self) -> aiohttp_client_cache.CachedSession:
        return self._cached_session

    @property
    def lib_version(self) -> Version:
        return VERSION

    @property
    def bot_id(self) -> str:
        return self._user_id

    @SingletonCallable.run_once_async
    async def initialize(self) -> None:
        try:
            if not self._initiated:
                await self._maybe_start_pylav()
        except Exception as exc:
            LOGGER.critical("Failed start up", exc_info=exc)
            raise exc

    async def _maybe_start_pylav(self):
        async with self._asyncio_lock:
            if not self._initiated:
                self._initiated = True
                self.ready.clear()
                await self._wait_until_ready()
                if IS_POSTGRES:
                    await DATABASE_ENGINE.start_connection_pool(max_size=100)
                (
                    spotify_client_id,
                    spotify_client_secret,
                    deezer_token,
                    yandex_access_token,
                    apple_music_token,
                ) = await self._get_service_tokens()

                self.bot.add_listener(self.on_pylav_red_api_tokens_update, name="on_red_api_tokens_update")
                if isinstance(self.bot, discord.AutoShardedClient):
                    self.bot.add_listener(self.on_pylav_shard_ready, name="on_shard_ready")
                else:
                    self.bot.add_listener(self.on_pylav_ready, name="on_ready")
                await self._initialise_modules()
                config_data = await self._config.fetch_all()
                java_path = config_data["java_path"]
                config_folder = config_data["config_folder"]
                localtrack_folder = config_data["localtrack_folder"]
                if java_path != JAVA_EXECUTABLE and os.path.exists(JAVA_EXECUTABLE):
                    await self._config.update_java_path(JAVA_EXECUTABLE)
                LOGGER.info("Lavalink folder: %s", LAVALINK_DOWNLOAD_DIR)
                LOGGER.info("Config folder: %s", config_folder)
                LOGGER.info("Localtracks folder: %s", localtrack_folder)
                self._config_folder = aiopath.AsyncPath(config_folder)
                localtrack_folder = aiopath.AsyncPath(localtrack_folder)
                bundled_node_config = self._node_config_manager.bundled_node_config()
                spotify_client_id, spotify_client_secret = await self._initialize_yaml_config(
                    bundled_node_config,
                    spotify_client_id,
                    spotify_client_secret,
                    deezer_token,
                    yandex_access_token,
                    apple_music_token,
                )
                self._has_deezer_support = True
                self._has_apm_support = True
                self._has_spotify_support = False
                self._has_yandex_support = False

                if spotify_client_id and spotify_client_secret:
                    self._spotify_auth = ClientCredentialsFlow(
                        client_id=spotify_client_id, client_secret=spotify_client_secret
                    )
                    self._has_spotify_support = True
                    await self.update_spotify_tokens(client_id=spotify_client_id, client_secret=spotify_client_secret)
                if yandex_access_token:
                    self._has_yandex_support = True
                    await self.update_yandex_tokens(token=yandex_access_token)
                if apple_music_token:
                    await self.update_applemusic_tokens(
                        token=apple_music_token, country_code=MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE
                    )
                await self._run_post_init_jobs(java_path, localtrack_folder)
                self.ready.set()
                LOGGER.info("PyLav is ready")

    async def _initialise_modules(self):
        await self._lib_config_manager.initialize()
        self._config = self._lib_config_manager.get_config()
        await RealNode.create_managed(identifier=self.bot.user.id)
        await self._update_schema_manager.run_updates()
        await self._radio_manager.initialize()
        await self._player_manager.initialize()
        await self.player_config_manager.initialize_global_config()

    async def _run_post_init_jobs(self, java_path, localtrack_folder) -> None:

        from pylav.players.query.local_files import LocalFile

        await LocalFile.add_root_folder(path=localtrack_folder, create=True)
        self._user_id = str(self._bot.user.id)
        self._local_node_manager = LocalNodeManager(self)
        enable_managed_node = await self.lib_db_manager.get_config().fetch_enable_managed_node()
        time_now = get_now_utc()
        await self._maybe_start_bundled_node(enable_managed_node, java_path)
        await self.node_manager.connect_to_all_nodes()
        await self.node_manager.wait_until_ready()
        await self._maybe_wait_until_bundled_node(enable_managed_node)
        await self._update_schema_manager.run_deferred_tasks_which_depend_on_node()
        await self._maybe_update_next_execution_bundled_playlist(time_now)
        await self._maybe_update_next_execution_bundled_external_playlists(time_now)
        await self._maybe_update_next_execution_external_playlists(time_now)
        await self._maybe_force_update_bundled_playlists()
        await self._add_scheduler_job_cache_cleanup()
        await self._add_scheduler_job_bundled_playlist()
        await self._add_scheduler_job_bundled_external_playlists()
        await self._add_scheduler_job_external_playlists()
        self._scheduler.start()
        await self.player_manager.restore_player_states()

    async def _maybe_start_bundled_node(self, enable_managed_node: bool, java_path: str) -> None:
        # noinspection PyProtectedMember
        if (
            not (self._node_manager._unmanaged_external_password and self._node_manager._unmanaged_external_host)
            and enable_managed_node
        ):
            await self._local_node_manager.start(java_path=java_path)
        else:
            self._local_node_manager.ready.set()

    async def _maybe_force_update_bundled_playlists(self) -> None:
        total_bundled_playlists = len(BUNDLED_PLAYLIST_IDS)
        if not self._has_deezer_support:
            total_bundled_playlists -= len(BUNDLED_DEEZER_PLAYLIST_IDS)
        if not self._has_spotify_support:
            total_bundled_playlists -= len(BUNDLED_SPOTIFY_PLAYLIST_IDS)

        if await self.playlist_db_manager.count() < total_bundled_playlists:
            time_now = get_now_utc()
            self._wait_for_playlists.clear()
            await self._config.update_next_execution_update_bundled_playlists(time_now - datetime.timedelta(days=7))
            await self._config.update_next_execution_update_bundled_external_playlists(
                time_now - datetime.timedelta(days=7)
            )

    async def _maybe_wait_until_bundled_node(self, enable_managed_node):
        # noinspection PyProtectedMember
        if (
            not (self._node_manager._unmanaged_external_password and self._node_manager._unmanaged_external_host)
            and enable_managed_node
        ):
            await self._local_node_manager.wait_until_connected()

    async def _wait_until_ready(self):
        if hasattr(self.bot, "wait_until_red_ready"):
            LOGGER.debug("Running on a Red bot, waiting for Red to be ready")
            await self.bot.wait_until_red_ready()
        else:
            LOGGER.debug("Running a discord.py bot waiting for bot to be ready")
            await self.bot.wait_until_ready()

    async def _get_service_tokens(self):
        if hasattr(self.bot, "get_shared_api_tokens") and callable(getattr(self.bot, "get_shared_api_tokens")):
            spotify = await self.bot.get_shared_api_tokens("spotify")
            client_id = spotify.get("client_id")
            client_secret = spotify.get("client_secret")
            if client_id and client_secret:
                LOGGER.debug("Existing Spotify tokens found; Using clientID - %s", client_id)
            else:
                client_id = MANAGED_NODE_SPOTIFY_CLIENT_ID
                client_secret = MANAGED_NODE_SPOTIFY_CLIENT_SECRET
            deezer = await self.bot.get_shared_api_tokens("deezer")
            deezer_token = deezer.get("master_token")
            if deezer_token:
                LOGGER.debug("Existing Deezer token found; Using it")
            else:
                deezer_token = MANAGED_NODE_DEEZER_KEY

            yandex = await self.bot.get_shared_api_tokens("yandexmusic")
            yandex_access_token = yandex.get("token")
            if yandex_access_token:
                LOGGER.debug("Existing Yandex Music token found; Using it")
            else:
                yandex_access_token = MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN

            apple_music = await self.bot.get_shared_api_tokens("apple_music")
            apple_music_token = apple_music.get("token")
            if apple_music_token:
                LOGGER.debug("Existing Apple Music token found; Using it")
            else:
                apple_music_token = MANAGED_NODE_APPLE_MUSIC_API_KEY
        else:
            LOGGER.info("PyLav being run from a non Red bot")
            client_id = MANAGED_NODE_SPOTIFY_CLIENT_ID
            client_secret = MANAGED_NODE_SPOTIFY_CLIENT_SECRET
            deezer_token = MANAGED_NODE_DEEZER_KEY
            yandex_access_token = MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN
            apple_music_token = MANAGED_NODE_APPLE_MUSIC_API_KEY

        return client_id, client_secret, deezer_token, yandex_access_token, apple_music_token

    async def _add_scheduler_job_external_playlists(self):
        next_execution_update_external_playlists = await self._config.fetch_next_execution_update_external_playlists()
        self._scheduler.add_job(
            self.playlist_db_manager.update_external_playlists,
            trigger="interval",
            days=TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
            max_instances=1,
            next_run_time=next_execution_update_external_playlists,
            replace_existing=True,
            name="update_external_playlists",
            coalesce=True,
            id=f"{self.bot.user.id}-update_external_playlists",
            misfire_grace_time=None,  # type: ignore
        )
        LOGGER.info(
            "Scheduling next run of External Playlist update task to: %s",
            next_execution_update_external_playlists,
        )

    async def _add_scheduler_job_bundled_external_playlists(self):
        next_execution_update_bundled_external_playlists = (
            await self._config.fetch_next_execution_update_bundled_external_playlists()
        )
        self._scheduler.add_job(
            self.playlist_db_manager.update_bundled_external_playlists,
            trigger="interval",
            days=TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
            max_instances=1,
            next_run_time=next_execution_update_bundled_external_playlists,
            replace_existing=True,
            name="update_bundled_external_playlists",
            coalesce=True,
            id=f"{self.bot.user.id}-update_bundled_external_playlists",
            misfire_grace_time=None,  # type: ignore
        )
        LOGGER.info(
            "Scheduling next run of Bundled External Playlist update task to: %s",
            next_execution_update_bundled_external_playlists,
        )

    async def _add_scheduler_job_bundled_playlist(self):
        next_execution_update_bundled_playlists = await self._config.fetch_next_execution_update_bundled_playlists()
        self._scheduler.add_job(
            self.playlist_db_manager.update_bundled_playlists,
            trigger="interval",
            days=TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
            max_instances=1,
            next_run_time=next_execution_update_bundled_playlists,
            replace_existing=True,
            name="update_bundled_playlists",
            coalesce=True,
            id=f"{self.bot.user.id}-update_bundled_playlists",
            misfire_grace_time=None,  # type: ignore
        )
        LOGGER.info(
            "Scheduling next run of Bundled Playlist update task to: %s",
            next_execution_update_bundled_playlists,
        )

    async def _add_scheduler_job_cache_cleanup(self):
        self._scheduler.add_job(
            self._query_cache_manager.delete_old,
            trigger="interval",
            seconds=600,
            max_instances=1,
            replace_existing=True,
            name="cache_delete_old",
            coalesce=True,
            id=f"{self.bot.user.id}-cache_delete_old",
        )

    async def _maybe_update_next_execution_external_playlists(self, time_now):
        if await self._config.fetch_next_execution_update_external_playlists() is None:
            await self._config.update_next_execution_update_external_playlists(
                time_now + datetime.timedelta(minutes=30)
            )

    async def _maybe_update_next_execution_bundled_external_playlists(self, time_now):
        if await self._config.fetch_next_execution_update_bundled_external_playlists() is None:
            await self._config.update_next_execution_update_bundled_external_playlists(
                time_now + datetime.timedelta(minutes=10, days=7)
            )

    async def _maybe_update_next_execution_bundled_playlist(self, time_now):
        if await self._config.fetch_next_execution_update_bundled_playlists() is None:
            await self._config.update_next_execution_update_bundled_playlists(
                time_now + datetime.timedelta(minutes=5, days=1)
            )

    @staticmethod
    async def _initialize_yaml_config(
        bundled_node_config,
        spotify_client_id,
        spotify_client_secret,
        deezer_token,
        yandex_access_token,
        apple_music_token,
    ):
        yaml_data = await bundled_node_config.fetch_yaml()
        need_update = False
        if not await asyncstdlib.all([spotify_client_id, spotify_client_secret]):
            yaml_data["plugins"]["lavasrc"]["sources"]["spotify"] = False
            need_update = True
        elif await asyncstdlib.all([spotify_client_id, spotify_client_secret]):
            yaml_data["plugins"]["lavasrc"]["spotify"]["clientId"] = spotify_client_id
            yaml_data["plugins"]["lavasrc"]["spotify"]["clientSecret"] = spotify_client_secret
            yaml_data["plugins"]["lavasrc"]["sources"]["spotify"] = True
            need_update = True
        if deezer_token:
            yaml_data["plugins"]["lavasrc"]["deezer"]["masterDecryptionKey"] = deezer_token
            yaml_data["plugins"]["lavasrc"]["sources"]["deezer"] = True
            need_update = True
        if yandex_access_token:
            yaml_data["plugins"]["lavasrc"]["yandexmusic"]["accessToken"] = yandex_access_token
            yaml_data["plugins"]["lavasrc"]["sources"]["yandexmusic"] = True
            need_update = True
        if apple_music_token:
            yaml_data["plugins"]["lavasrc"]["applemusic"]["mediaAPIToken"] = apple_music_token
            yaml_data["plugins"]["lavasrc"]["sources"]["applemusic"] = True
        if need_update:
            await bundled_node_config.update_yaml(yaml_data)
        return spotify_client_id, spotify_client_secret

    async def register(self, cog: DISCORD_COG_TYPE) -> None:
        LOGGER.debug("Registering cog %s", cog.__cog_name__)
        if (instance := getattr(self.bot, "pylav", None)) and not isinstance(instance, Client):
            raise AnotherClientAlreadyRegisteredException(
                f"Another client instance has already been registered to bot.pylav with type: {type(instance)}"
            )
        self.__cogs_registered.add(cog.__cog_name__)

    async def update_spotify_tokens(self, client_id: str, client_secret: str, **kwargs) -> None:
        LOGGER.info("Updating Spotify Tokens")
        LOGGER.debug("New Spotify token: ClientId: %s || ClientSecret: %s", client_id, client_secret)
        self._spotify_auth = ClientCredentialsFlow(client_id=client_id, client_secret=client_secret)
        bundled_node_config = self._node_config_manager.bundled_node_config()
        bundled_node_config_yaml = await bundled_node_config.fetch_yaml()
        bundled_node_config_yaml["plugins"]["lavasrc"]["sources"]["spotify"] = True
        bundled_node_config_yaml["plugins"]["lavasrc"]["spotify"]["clientId"] = client_id
        bundled_node_config_yaml["plugins"]["lavasrc"]["spotify"]["clientSecret"] = client_secret
        bundled_node_config_yaml["plugins"]["lavasrc"]["spotify"]["countryCode"] = kwargs.get(
            "country_code", MANAGED_NODE_SPOTIFY_COUNTRY_CODE
        )
        await bundled_node_config.update_yaml(bundled_node_config_yaml)

    async def update_deezer_tokens(self, master_token: str, **kwargs: Any) -> None:
        LOGGER.info("Updating Deezer Tokens")
        LOGGER.debug("New Deezer token: %s", master_token)
        bundled_node_config = self._node_config_manager.bundled_node_config()
        bundled_node_config_yaml = await bundled_node_config.fetch_yaml()
        bundled_node_config_yaml["plugins"]["lavasrc"]["sources"]["deezer"] = True
        bundled_node_config_yaml["plugins"]["lavasrc"]["deezer"]["masterDecryptionKey"] = master_token
        await bundled_node_config.update_yaml(bundled_node_config_yaml)

    async def update_yandex_tokens(self, token: str, **kwargs: Any) -> None:
        LOGGER.info("Updating Yandex Tokens")
        LOGGER.debug("New Yandex token: %s", token)
        bundled_node_config = self._node_config_manager.bundled_node_config()
        bundled_node_config_yaml = await bundled_node_config.fetch_yaml()
        bundled_node_config_yaml["plugins"]["lavasrc"]["sources"]["yandexmusic"] = True
        bundled_node_config_yaml["plugins"]["lavasrc"]["yandexmusic"]["accessToken"] = token
        await bundled_node_config.update_yaml(bundled_node_config_yaml)

    async def update_applemusic_tokens(self, token: str, country_code: str, **kwargs: Any) -> None:
        LOGGER.info("Updating Apple Music Tokens")
        LOGGER.debug("New Apple Music tokens: mediaAPIToken %s || countryCode %s", token, country_code)
        bundled_node_config = self._node_config_manager.bundled_node_config()
        bundled_node_config_yaml = await bundled_node_config.fetch_yaml()
        bundled_node_config_yaml["plugins"]["lavasrc"]["sources"]["applemusic"] = True
        bundled_node_config_yaml["plugins"]["lavasrc"]["applemusic"]["mediaAPIToken"] = token
        bundled_node_config_yaml["plugins"]["lavasrc"]["applemusic"]["countryCode"] = kwargs.get(
            "country_code", MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE
        )
        await bundled_node_config.update_yaml(bundled_node_config_yaml)

    async def update_google_account(self, email: str, password: str, **kwargs: Any) -> None:
        LOGGER.info("Updating Google Account")
        LOGGER.debug("New Google Account: %s", email)
        bundled_node_config = self._node_config_manager.bundled_node_config()
        bundled_node_config_yaml = await bundled_node_config.fetch_yaml()
        bundled_node_config_yaml["lavalink"]["server"]["youtubeConfig"]["email"] = email
        bundled_node_config_yaml["lavalink"]["server"]["youtubeConfig"]["password"] = password
        await bundled_node_config.update_yaml(bundled_node_config_yaml)

    async def add_node(
        self,
        *,
        unique_identifier: int,
        host: str,
        port: int,
        password: str,
        resume_timeout: int = 60,
        name: str = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        managed: bool = False,
        yaml: dict | None = None,
        disabled_sources: list[str] = None,
        extras: dict = None,
        temporary: bool = False,
    ) -> Node:
        """
        Adds a node to PyLav's node manager.

        Parameters
        ----------
        host: :class:`str`
            The address of the Lavalink node.
        port: :class:`int`
            The port to use for websocket and REST connections.
        password: :class:`str`
            The password used for authentication.
        resume_timeout: Optional[:class:`int`]
            How long the node should wait for a connection while disconnected before clearing all players.
            Defaults to `60`.
        name: :class:`str`
            An identifier for the node that will show in logs. Defaults to `None`
        reconnect_attempts: Optional[:class:`int`]
            The amount of times connection with the node will be reattempted before giving up.
            Set to `-1` for infinite. Defaults to `3`.
        ssl: Optional[:class:`bool`]
            Whether to use SSL for the connection. Defaults to `False`.
        search_only: :class:`bool`
            Whether the node should only be used for searching. Defaults to `False`.
        unique_identifier: :class:`in`
            A unique identifier for the node. Defaults to `None`.
        yaml: Optional[:class:`dict`]
            A dictionary of extra information to be stored in the node. Defaults to `None`.
        extras: Optional[:class:`dict`]
            A dictionary of extra information to be stored in the node. Defaults to `None`.
        managed: :class:`bool`
            Whether the node is managed by the client. Defaults to `False`.
        disabled_sources: Optional[:class:`list`[:class:`str`]]
            A list of sources that should be disabled for the node. Defaults to `None`.
        temporary: :class:`bool`
            Whether the node is temporary. Defaults to `False`.
        """
        return await self.node_manager.add_node(
            host=host,
            port=port,
            password=password,
            resume_timeout=resume_timeout,
            name=name,
            reconnect_attempts=reconnect_attempts,
            ssl=ssl,
            search_only=search_only,
            unique_identifier=unique_identifier,
            managed=managed,
            yaml=yaml,
            disabled_sources=disabled_sources,
            extras=extras or {},
            temporary=temporary,
        )

    async def decode_track(
        self, track: str, feature: str = None, raise_on_failure: bool = False, lazy: bool = False
    ) -> TrackObject | HTTPException:
        """|coro|
        Decodes a base64-encoded track string into a dict.

        Parameters
        ----------
        track: :class:`str`
            The base64-encoded `track` string.
        feature: Optional[:class:`str`]
            The feature to decode the track for. Defaults to `None`.
        raise_on_failure: Optional[:class:`bool`]
            Whether to raise an exception if the track fails to decode. Defaults to `False`.
        lazy: :class:`bool`
            Weather to decode within the Bot or send to Lavalink. Defaults to `False`.

        Returns
        -------
        LavalinkTrackObject
            An object representing the track's information.
        """
        if lazy:
            with contextlib.suppress(Exception):
                return await async_decoder(track)
        if not self.node_manager.available_nodes:
            raise NoNodeAvailableException(_("No available nodes!"))
        node = await self.node_manager.find_best_node(feature=feature)
        if node is None and feature:
            raise NoNodeWithRequestFunctionalityAvailableException(
                _("No node with {feature} functionality available!").format(feature=feature), feature=feature
            )
        try:
            response = await node.fetch_decodetrack(track, raise_on_failure=raise_on_failure)
            if isinstance(response, TrackObject):
                return response
            raise TypeError
        except Exception as exc:  # noqa
            return await async_decoder(track)

    async def decode_tracks(
        self, tracks: list, feature: str = None, raise_on_failure: bool = False
    ) -> list[TrackObject]:
        """|coro|
        Decodes a list of base64-encoded track strings into a dict.

        Parameters
        ----------
        tracks: list[:class:`str`]
            A list of base64-encoded `track` strings.
        feature: Optional[:class:`str`]
            The feature to decode the tracks for. Defaults to `None`.
        raise_on_failure: Optional[:class:`bool`]
            Whether to raise an exception if the tracks fail to decode. Defaults to `False`.

        Returns
        -------
        List[LavalinkTrackObject]
            A list of LavalinkTrackObject representing track information.
        """
        if not self.node_manager.available_nodes:
            raise NoNodeAvailableException(_("No available nodes!"))
        node = await self.node_manager.find_best_node(feature=feature)
        if node is None and feature:
            raise NoNodeWithRequestFunctionalityAvailableException(
                _("No node with {feature} functionality available!").format(feature=feature), feature=feature
            )
        try:
            response = await node.post_decodetracks(tracks, raise_on_failure=raise_on_failure)
            if isinstance(response, HTTPException):
                raise TypeError
            return response
        except Exception:  # noqa
            response_tracks = []
            async for track in AsyncIter(tracks):
                with contextlib.suppress(Exception):
                    response_tracks.append(await async_decoder(track))
            return response_tracks

    @staticmethod
    async def routeplanner_status(node: Node) -> RoutePlannerStatus:
        """|coro|
        Gets the route-planner status of the target node.

        Parameters
        ----------
        node: :class:`Node`
            The node to use for the query.

        Returns
        -------
        RoutePlannerStatusResponseObject
            An object representing the route-planner information.
        """
        return await node.fetch_routeplanner_status()

    @staticmethod
    async def routeplanner_free_address(node: Node, address: str) -> None:
        """|coro|
        Gets the route-planner status of the target node.

        Parameters
        ----------
        node: :class:`Node`
            The node to use for the query.
        address: :class:`str`
            The address to free.
        """
        return await node.post_routeplanner_free_address(address)

    @staticmethod
    async def routeplanner_free_all_failing(node: Node) -> None:
        """|coro|
        Gets the route-planner status of the target node.

        Parameters
        ----------
        node: :class:`Node`
            The node to use for the query.
        """
        return await node.post_routeplanner_free_all()

    def dispatch_event(self, event: PyLavEvent) -> None:
        asyncio.create_task(self._dispatch_event(event))

    async def _dispatch_event(self, event: PyLavEvent) -> None:
        """|coro|
        Dispatches the given event to all registered hooks.

        Parameters
        ----------
        event: :class:`Event`
            The event to dispatch to the hooks.
        """
        event_dispatcher = [self._dispatch_manager.dispatch]

        task_list = []
        async for hook in itertools.chain(
            event_dispatcher,
        ):
            task = asyncio.create_task(hook(event))  # type: ignore
            task.set_name(f"Event hook {hook.__name__}")
            task.add_done_callback(self.__done_callback)
            task_list.append(task)
        await asyncio.gather(*task_list)

    @staticmethod
    def __done_callback(task: asyncio.Task) -> None:
        if (exc := task.exception()) is not None:
            name = task.get_name()
            LOGGER.warning("Event hook %s encountered an exception!", name)
            LOGGER.debug("Event hook %s encountered an exception!", name, exc_info=exc)

    async def unregister(self, cog: discord.ext.commands.Cog):
        """|coro|
        Unregister the specified Cog and if no cogs are left closes the client.

        Parameters
        ----------
        cog: :class:`discord.ext.commands.Cog`
            The cog to unregister.
        """
        if self._shutting_down:
            return
        async with self._asyncio_lock:
            if not self._shutting_down:
                self.__cogs_registered.discard(cog.__cog_name__)
                LOGGER.debug("%s has been unregistered", cog.__cog_name__)
                if not self.__cogs_registered:
                    self.bot.remove_listener(self.on_pylav_red_api_tokens_update, name="on_red_api_tokens_update")
                    self.bot.remove_listener(self.on_pylav_shard_resumed, name="on_shard_resumed")
                    self.bot.remove_listener(self.on_pylav_shard_ready, name="on_shard_ready")
                    self.bot.remove_listener(self.on_pylav_ready, name="on_ready")
                    self.bot.remove_listener(self.on_pylav_resumed, name="on_resumed")
                    self._shutting_down = True
                    self.ready.clear()
                    try:
                        Client._instances.clear()
                        SingletonCallable.reset()
                        self._initiated = False
                        await self.player_manager.save_all_players()
                        await self.player_manager.shutdown()
                        await self._node_manager.close()
                        await self._local_node_manager.shutdown()
                        await self._session.close()
                        await self._cached_session.close()

                        if self._scheduler:
                            with contextlib.suppress(Exception):
                                self._scheduler.shutdown(wait=True)
                    except Exception as e:
                        LOGGER.critical("Failed to shutdown the client", exc_info=e)
                    if self.__old_process_command_method is not None:
                        self.bot.process_commands = self.__old_process_command_method
                    if self.__old_get_context is not None:
                        self.bot.get_context = self.__old_get_context
                    del self.bot._pylav_client  # noqa
                    await DATABASE_ENGINE.close_connection_pool()
                    LOGGER.info("All cogs have been unregistered, PyLav client has been shutdown")

    def get_player(self, guild: discord.Guild | int | None) -> Player | None:
        """Gets the player for the target guild.

        Parameters
        ----------
        guild: :class:`discord.Guild`
            The guild to get the player for.

        Returns
        -------
        :class:`Player`
            The player for the target guild.
        """
        if not guild:
            return None
        if not isinstance(guild, int):
            guild = guild.id
        return self.player_manager.get(guild)

    async def connect_player(
        self,
        requester: discord.Member,
        channel: discord.channel.VocalGuildChannel,
        node: Node = None,
        self_deaf: bool = True,
    ) -> Player:
        """|coro|
        Connects the player for the target guild.

        Parameters
        ----------
        channel: :class:`discord.channel.VocalGuildChannel`
            The channel to connect to.
        node: :class:`Node`
            The node to use for the connection.
        self_deaf: :class:`bool`
            Whether the bot should be deafened.
        requester: :class:`discord.Member`
            The member requesting the connection.
        Returns
        -------
        :class:`Player`
            The player for the target guild.
        """
        return await self.player_manager.create(channel, channel.rtc_region, node, self_deaf, requester)

    async def construct_embed(
        self,
        *,
        embed: discord.Embed = None,
        colour: discord.Colour | int | None = None,
        color: discord.Colour | int | None = None,
        title: str = None,
        embed_type: EmbedType = "rich",
        url: str = None,
        description: str = None,
        timestamp: datetime.datetime = None,
        author_name: str = None,
        author_url: str = None,
        thumbnail: str = None,
        footer: str = None,
        footer_url: str = None,
        messageable: Messageable | DISCORD_INTERACTION_TYPE = None,
    ) -> discord.Embed:

        if messageable and not colour and not color and hasattr(self._bot, "get_embed_color"):
            colour = await self._bot.get_embed_color(messageable)
        elif colour or color:
            colour = colour or color
        if timestamp and isinstance(timestamp, datetime.datetime):
            timestamp = timestamp
        else:
            timestamp = get_now_utc()
        contents = dict(
            title=title,
            type=embed_type,
            url=url,
            description=description,
            timestamp=timestamp.isoformat(),
        )
        embed = embed.to_dict() if embed is not None else {}
        contents |= embed
        new_embed = discord.Embed.from_dict(contents)
        new_embed.color = colour

        if footer:
            new_embed.set_footer(text=footer, icon_url=footer_url)
        if thumbnail:
            new_embed.set_thumbnail(url=thumbnail)
        if author_url and author_name:
            new_embed.set_author(name=author_name, icon_url=author_url)
        return new_embed

    async def get_context(
        self, what: discord.Message | DISCORD_CONTEXT_TYPE | DISCORD_INTERACTION_TYPE
    ) -> PyLavContext:
        if isinstance(what, PyLavContext):
            return what
        elif isinstance(what, Context):
            ctx_ = what.interaction or what.message
            ctx: PyLavContext = await self._bot.get_context(ctx_, cls=PyLavContext)  # type: ignore

        else:
            ctx: PyLavContext = await self._bot.get_context(what, cls=PyLavContext)  # type: ignore
        return ctx

    async def update_localtracks_folder(self, folder: str | None) -> None:

        localtrack_folder = aiopath.AsyncPath(folder) if folder else self._config_folder / "music"

        from pylav.players.query.local_files import LocalFile

        await LocalFile.add_root_folder(path=localtrack_folder, create=True)

    async def get_all_players(self) -> AsyncIterator[Player]:

        return asyncstdlib.iter(self.player_manager)

    async def get_managed_node(self) -> Node | None:
        available_nodes = await asyncstdlib.list(
            asyncstdlib.filter(operator.attrgetter("available"), self.node_manager.managed_nodes)
        )

        return random.choice(available_nodes) if available_nodes else None

    async def get_my_node(self) -> Node | None:
        return await asyncstdlib.anext(
            asyncstdlib.filter(lambda n: n.identifier == self.bot.user.id, self.node_manager.managed_nodes),
            default=None,  # type: ignore
        )

    async def _get_tracks(
        self,
        query: Query,
        first: bool = False,
        bypass_cache: bool = False,
        player: Player | None = None,
    ) -> LoadTrackResponses:
        """|coro|
        Gets all tracks associated with the given query.

        Parameters
        ----------
        query: :class:`Query`
            The query to perform a search for
        first: Optional[:class:`bool`]
            Whether to only return the first track. Defaults to `False`.
        bypass_cache: Optional[:class:`bool`]
            Whether to bypass the cache. Defaults to `False`.

        Returns
        -------
        :class:`LavalinkLoadTrackObjects`
            A LavalinkLoadTrackObjects representing Lavalink response.
        """
        if not self.node_manager.available_nodes:
            raise NoNodeAvailableException("No available nodes!")
        node = await self.node_manager.find_best_node(
            feature=query.requires_capability,
            region=player.region if player else None,
            coordinates=player.coordinates if player else None,
        )
        if node is None:
            raise NoNodeWithRequestFunctionalityAvailableException(
                _("No node with {query.requires_capability} functionality available!").format(query=query),
                query.requires_capability,
            )
        return await node.get_track(query, first=first, bypass_cache=bypass_cache)

    async def get_all_tracks_for_queries(
        self,
        *queries: Query,
        requester: discord.Member,
        player: Player | None = None,
        bypass_cache: bool = False,
        enqueue: bool = True,
        partial: bool = PREFER_PARTIAL_TRACKS,
    ) -> tuple[list[Track], int, list[Query]]:  # sourcery no-metrics
        """High level interface to get and return all tracks for a list of queries.

        This will automatically handle playlists, albums, searches and local files.

        Parameters
        ----------
        queries : `Query`
            The list of queries to search for.
        bypass_cache : `bool`, optional
            Whether to bypass the cache and force a new search.
            Local files will always be bypassed.
        requester : `discord.Member`
            The user who requested the op.
        player : `Player`
            The player requesting the op.
        enqueue : `bool`, optional
            Whether to enqueue the tracks as needed
            while try are processed so users dont sit waiting for the bot to finish.
        partial : `bool`, optional
            Whether to return partial results

        Returns
        -------
        tracks : `List[Track]`
            The list of tracks found.
        total_tracks : `int`
            The total number of tracks found.
        queries : `List[Query]`
            The list of queries that were not found.

        """
        successful_tracks = []
        queries_failed = []
        track_count = 0
        for query in queries:
            track_count = await self._get_tracks_single_query(
                bypass_cache,
                enqueue,
                player,
                queries_failed,
                query,
                requester,
                successful_tracks,
                track_count,
                query.is_partial if partial is None else partial,
            )
        return successful_tracks, track_count, queries_failed

    async def _get_tracks_single_query(
        self, bypass_cache, enqueue, player, queries_failed, query, requester, successful_tracks, track_count, partial
    ):
        async for sub_query in self._yield_recursive_queries(query):
            node = await self.node_manager.find_best_node(
                region=player.region if player else None,
                coordinates=player.coordinates if player else None,
                feature=sub_query.requires_capability,
            )
            if node is None:
                queries_failed.append(sub_query)
                continue
            await self._get_tracks_play_or_enqueue(enqueue, player, requester, successful_tracks)
            if partial and sub_query.is_single:
                track_count = await self._get_tracks_partial_single(
                    enqueue, node, player, requester, sub_query, successful_tracks, track_count
                )
            elif sub_query.is_search or sub_query.is_single:
                track_count = await self._get_tracks_search_or_single(
                    bypass_cache,
                    node,
                    player,
                    queries_failed,
                    requester,
                    sub_query,
                    successful_tracks,
                    track_count,
                    partial,
                )
            elif (
                (sub_query.is_playlist or sub_query.is_album)
                and not sub_query.is_local
                and not sub_query.is_custom_playlist
            ):
                track_count = await self._get_tracks_playlist_or_album_no_local(
                    bypass_cache,
                    enqueue,
                    node,
                    player,
                    queries_failed,
                    requester,
                    sub_query,
                    successful_tracks,
                    track_count,
                )
            elif (sub_query.is_local or sub_query.is_custom_playlist) and sub_query.is_album:
                track_count = await self._get_tracks_local_album(
                    enqueue, node, player, queries_failed, requester, sub_query, successful_tracks, track_count, partial
                )
            else:
                queries_failed.append(sub_query)
                LOGGER.warning("Unhandled query: %s, %s", sub_query.to_dict(), sub_query.query_identifier)
        return track_count

    async def _get_tracks_partial_single(
        self, enqueue, node, player, requester, sub_query, successful_tracks, track_count
    ):
        track_count += 1
        if cached_query := await self.node_manager.client.query_cache_manager.fetch_query(sub_query):
            data = (await cached_query.fetch_tracks())[0]
        else:
            data = None
        successful_tracks.append(
            await Track.build_track(
                data=data,
                node=node,
                query=sub_query,
                requester=requester.id,
            )
        )
        if enqueue and successful_tracks and not player.is_playing and not player.paused:
            track = successful_tracks.pop()
            await player.play(track, await track.query(), requester)
        return track_count

    @staticmethod
    async def _get_tracks_play_or_enqueue(enqueue, player, requester, successful_tracks):
        # Query tracks as the queue builds as this may be a slow operation
        if enqueue and successful_tracks and not player.is_playing and not player.paused:
            track = successful_tracks.pop()
            await player.play(track, await track.query(), requester)
        elif successful_tracks and player.is_playing and player.queue.empty():
            track = successful_tracks.pop()
            await player.add(requester.id, track, query=await track.query())

    async def _get_tracks_local_album(
        self, enqueue, node, player, queries_failed, requester, sub_query, successful_tracks, track_count, partial
    ):
        yielded = False
        async for local_track in sub_query.get_all_tracks_in_folder():
            yielded = True
            if partial:
                track_count += 1
                successful_tracks.append(
                    await Track.build_track(
                        data=MISSING,
                        node=node,
                        query=local_track,
                        requester=requester.id,
                    )
                )
                continue
            response = await self._get_tracks(player=player, query=local_track, first=True, bypass_cache=True)
            if __ := response.tracks[0].encoded:
                track_count += 1
                successful_tracks.append(
                    await Track.build_track(
                        data=response.tracks[0],
                        node=node,
                        query=await Query.from_string(response.tracks[0].info.uri),
                        requester=requester.id,
                    )
                )
                # Query tracks as the queue builds as this may be a slow operation
                if enqueue and successful_tracks and not player.is_playing and not player.paused:
                    track = successful_tracks.pop()
                    await player.play(track, await track.query(), requester)
        if not yielded:
            queries_failed.append(sub_query)
        return track_count

    async def _get_tracks_playlist_or_album_no_local(
        self,
        bypass_cache,
        enqueue,
        node,
        player,
        queries_failed,
        requester,
        sub_query,
        successful_tracks,
        track_count,
    ):
        response = await self._get_tracks(player=player, query=sub_query, bypass_cache=bypass_cache)
        track_list = response.tracks
        if not track_list:
            queries_failed.append(sub_query)
        for track in track_list:
            if __ := track.encoded:
                track_count += 1
                successful_tracks.append(
                    await Track.build_track(
                        data=track,
                        node=node,
                        query=await Query.from_string(track.info.uri),
                        requester=requester.id,
                    )
                )
                # Query tracks as the queue builds as this may be a slow operation
                if enqueue and successful_tracks and not player.is_playing and not player.paused:
                    track = successful_tracks.pop()
                    await player.play(track, await track.query(), requester)
        return track_count

    async def _get_tracks_search_or_single(
        self, bypass_cache, node, player, queries_failed, requester, sub_query, successful_tracks, track_count, partial
    ):
        response = await self._get_tracks(player=player, query=sub_query, first=True, bypass_cache=bypass_cache)
        track_b64 = response.tracks[0].encoded
        if not track_b64:
            queries_failed.append(sub_query)
        if track_b64:
            track_count += 1
            new_query = await Query.from_string(response.tracks[0].info.uri)
            new_query.merge(sub_query, start_time=True)
            successful_tracks.append(
                await Track.build_track(
                    data=response.tracks[0],
                    node=node,
                    query=new_query,
                    requester=requester.id,
                )
            )
        return track_count

    @staticmethod
    async def _yield_recursive_queries(query: Query, recursion_depth: int = 0) -> AsyncIterator[Query]:
        """|coro|
        Gets all queries associated with the given query.
        Parameters
        ----------
        query: :class:`Query`
            The query to perform a search for
        """
        if query.invalid or recursion_depth > MAX_RECURSION_DEPTH:
            return
        recursion_depth += 1
        if query.is_m3u:
            # noinspection PyProtectedMember
            async for m3u in query._yield_m3u_tracks():
                with contextlib.suppress(Exception):
                    # noinspection PyProtectedMember
                    async for q in query._yield_tracks_recursively(m3u, recursion_depth):
                        yield q
        elif query.is_pylav:
            # noinspection PyProtectedMember
            async for pylav in query._yield_pylav_file_tracks():
                with contextlib.suppress(Exception):
                    # noinspection PyProtectedMember
                    async for q in query._yield_tracks_recursively(pylav, recursion_depth):
                        yield q
        elif query.is_pls:
            # noinspection PyProtectedMember
            async for pls in query._yield_pls_tracks():
                with contextlib.suppress(Exception):
                    # noinspection PyProtectedMember
                    async for q in query._yield_tracks_recursively(pls, recursion_depth):
                        yield q
        elif query.is_local and query.is_album:
            # noinspection PyProtectedMember
            async for local in query._yield_local_tracks():
                yield local
        else:
            yield query

    async def get_tracks(
        self,
        *queries: Query,
        bypass_cache: bool = False,
        fullsearch: bool = False,
        region: str | None = None,
        player: Player | None = None,
    ) -> LoadTrackResponses:
        """This method can be rather slow as it recursively queries all queries and their associated entries.

        Thus, if you are processing user input  you may be interested in using
        the :meth:`get_all_tracks_for_queries` where it can enqueue tracks as needed to the player.


        Parameters
        ----------
        queries : `Query`
            The list of queries to search for.
        bypass_cache : `bool`, optional
            Whether to bypass the cache and force a new search.
            Local files will always be bypassed.
        fullsearch : `bool`, optional
            if a Search query is passed wether to returrn a list of tracks instead of the first.
        region : `str`, optional
            The region to search in.
        player : `Player`, optional
            The player to use for enqueuing tracks.
        """
        output_tracks = []
        playlist_name = ""
        plugin_info = {}

        if region is None and player is not None:
            region = player.region
        if region is None:
            region = "us_east"
        for query in queries:
            async for subquery in self._yield_recursive_queries(query):
                response = await self.search_query(
                    subquery, bypass_cache=bypass_cache, fullsearch=fullsearch, region=region
                )
                if response is None or not response.tracks:
                    continue
                if subquery.is_playlist or subquery.is_album:
                    playlist_name = response.playlistInfo.name if response.playlistInfo else ""
                    output_tracks.extend(response.tracks)
                    plugin_info |= response.pluginInfo.to_dict()
                elif fullsearch and subquery.is_search or subquery.is_single:
                    output_tracks.extend(response.tracks)
                else:
                    LOGGER.error("Unknown query type: %s", subquery)
        data = {
            "playlistInfo": {
                "name": playlist_name if len(queries) == 1 else "",
                "selectedTrack": -1,
            },
            "pluginInfo": plugin_info,
            "loadType": "PLAYLIST_LOADED"
            if playlist_name and len(queries) == 1
            else "SEARCH_RESULT"
            if output_tracks
            else "LOAD_FAILED",
            "tracks": output_tracks,
        }
        if data["loadType"] == "LOAD_FAILED":
            data["exception"] = {"cause": "No tracks returned", "severity": "COMMON", "message": "No tracks found"}
        node = await self.node_manager.find_best_node()
        while not node:
            await asyncio.sleep(0.1)
            node = await self.node_manager.find_best_node()
        return node.parse_loadtrack_response(data)

    async def search_query(
        self, query: Query, bypass_cache: bool = False, fullsearch: bool = False, region: str | None = None
    ) -> LoadTrackResponses | None:
        """
        Search for the specified query returns a LoadTrackResponse object


        Parameters
        ----------
        query : `Query`
            The list of queries to search for.
        bypass_cache : `bool`, optional
            Whether to bypass the cache and force a new search.
            Local files will always be bypassed.
        fullsearch : `bool`, optional
            if a Search query is passed wether to returrn a list of tracks instead of the first.
        region : `str`, optional
            The region to search in.
        player : `Player`, optional
            The player to use for enqueuing tracks.
        """
        node = await self.node_manager.find_best_node(region=region, feature=query.requires_capability)
        if node is None:
            return
        if query.is_playlist or query.is_album or (fullsearch and query.is_search):
            return await node.get_track(query, bypass_cache=bypass_cache)
        elif query.is_single:
            return await node.get_track(query, first=True, bypass_cache=bypass_cache)
        else:
            LOGGER.error("Unknown query type: %s", query)

    async def remove_node(self, node_id: int):
        """Removes a node from the node manager"""
        if node := self.node_manager.get_node_by_id(node_id):
            await self.node_manager.remove_node(node)

    async def is_dj(
        self,
        user: discord.Member,
        guild: discord.Guild,
        *,
        additional_role_ids: list = None,
        additional_user_ids: list = None,
        bot: DISCORD_BOT_TYPE = None,
    ) -> bool:
        if additional_user_ids and user.id in additional_user_ids:
            return True
        if additional_role_ids and await asyncstdlib.any(r.id in additional_role_ids for r in user.roles):
            return True
        return await self.player_config_manager.is_dj(
            user=user, guild=guild, additional_role_ids=None, additional_user_ids=None, bot=bot
        )

    @staticmethod
    async def generate_mix_playlist(
        *,
        video_id: str | None = None,
        user_id: str | None,
        playlist_id: str | None = None,
        channel_id: str | None = None,
    ) -> str:
        if not await asyncstdlib.any([video_id, playlist_id, channel_id, user_id]):
            raise PyLavInvalidArgumentsException(
                _("To generate a mix playlist a Video, User, Channel or Playlist ID is necessary")
            )

        if await asyncstdlib.sum(1 for i in [video_id, playlist_id, channel_id, user_id] if i) > 1:
            raise PyLavInvalidArgumentsException(
                _(
                    "To generate a mix playlist a Video, User, Channel or Playlist ID is necessary. "
                    "However, you provided multiple"
                )
            )

        if video_id:
            return f"https://www.youtube.com/watch?list=RD{video_id}"
        if user_id:
            return f"https://www.youtube.com/watch?list=UU{user_id}"
        if playlist_id:
            return f"https://www.youtube.com/watch?list=RDAMPL{playlist_id}"
        if channel_id:
            return f"https://www.youtube.com/watch?list=RDCM{channel_id}"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import asyncstdlib
import discord
from discord import Emoji, PartialEmoji
from redbot.core.i18n import Translator

from pylav.helpers import emojis
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

if TYPE_CHECKING:
    from pylav.extension.red.ui.menus.nodes import AddNodeFlow, NodeManagerMenu


LOGGER = getLogger("PyLav.ext.red.ui.button.nodes")


_ = Translator("PyLav", Path(__file__))


class SSLNodeToggleButton(discord.ui.Button):
    view: AddNodeFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        emoji: str | Emoji | PartialEmoji,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        self.view.ssl = not self.view.ssl
        if self.view.ssl:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("Connecting to the node with SSL enabled")
                ),
                ephemeral=True,
            )
        else:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("Connecting to the node with SSL disabled")
                ),
                ephemeral=True,
            )


class SearchOnlyNodeToggleButton(discord.ui.Button):
    view: AddNodeFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        emoji: str | Emoji | PartialEmoji,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        self.view.search_only = not self.view.search_only
        if self.view.search_only:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("This node will only be used for searches")
                ),
                ephemeral=True,
            )
        else:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("This node will be used for search and playback")
                ),
                ephemeral=True,
            )


class AddNodeDoneButton(discord.ui.Button):
    view: AddNodeFlow

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.CHECK,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        if not await asyncstdlib.all([self.view.name, self.view.host, self.view.port, self.view.password]):
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("Please fill out all the fields before continuing")
                ),
                ephemeral=True,
            )
            return
        self.view.last_interaction = interaction
        self.view.done = True
        self.view.disabled_sources = self.view.disabled_sources_selector.values
        self.view.cancelled = False
        self.view.stop()
        await self.view.on_timeout()


class NodeButton(discord.ui.Button):
    view: AddNodeFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        op: Literal["name", "host", "port", "password", "timeout"],
        label: str = None,
        emoji: str | Emoji | PartialEmoji = None,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emoji,
            label=label,
            row=row,
        )
        self.cog = cog
        self.op = op

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            return await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        self.view.cancelled = False
        if self.op == "name":
            await self.view.prompt_name(interaction)
        elif self.op == "host":
            await self.view.prompt_host(interaction)
        elif self.op == "port":
            await self.view.prompt_port(interaction)
        elif self.op == "password":
            await self.view.prompt_password(interaction)
        elif self.op == "timeout":
            await self.view.prompt_resume_timeout(interaction)


class NodeDeleteButton(discord.ui.Button):
    view: NodeManagerMenu

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.TRASH,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        self.view.cancelled = False
        self.view.delete = not self.view.delete
        if self.view.delete:
            response = _("When you press done this node will be permanently delete")
        else:
            response = _("This node will no longer be deleted once you press done")

        await context.send(
            embed=await self.cog.pylav.construct_embed(messageable=interaction, description=response),
            ephemeral=True,
        )


class NodeShowEnabledSourcesButton(discord.ui.Button):
    view: NodeManagerMenu

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.CLOUD_SERVER,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if self.view.author.id != interaction.user.id:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        await context.send(
            embed=await self.cog.pylav.construct_embed(
                messageable=interaction,
                description=_("__Enabled sources__:\n{sources}").format(
                    sources="\n".join(map(str.title, self.view.source.target.capabilities))
                ),
            ),
            ephemeral=True,
        )



from __future__ import annotations

import datetime

JSON_TYPE = dict[str, "JSON_TYPE"] | list["JSON_TYPE"] | str | int | float | bool | None
JSON_DICT_TYPE = dict[str, JSON_TYPE]

JSON_WITH_DATE_TYPE = (
    dict[str, "JSON_WITH_DATE_TYPE"] | list["JSON_WITH_DATE_TYPE"] | str | int | float | bool | None | datetime.datetime
)
JSON_DICT_WITH_DATE_TYPE = dict[str, JSON_WITH_DATE_TYPE]



from __future__ import annotations

from pathlib import Path

import asyncstdlib
import discord
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import humanize_number
from redbot.vendored.discord.ext import menus

from pylav.extension.red.ui.menus.generic import BaseMenu
from pylav.extension.red.utils import rgetattr
from pylav.logging import getLogger
from pylav.players.player import Player
from pylav.type_hints.bot import DISCORD_COG_TYPE

LOGGER = getLogger("PyLav.ext.red.ui.sources.player")
_ = Translator("PyLav", Path(__file__))


class PlayersSource(menus.ListPageSource):
    def __init__(self, cog: DISCORD_COG_TYPE, specified_guild: int = None):
        super().__init__([], per_page=1)
        self.cog = cog
        self.current_player = None
        self.specified_guild = specified_guild

    @property
    def entries(self) -> list[Player]:
        if self.specified_guild is not None and (player := self.cog.pylav.player_manager.get(self.specified_guild)):
            return [player]
        return self.cog.pylav.player_manager.connected_players

    @entries.setter
    def entries(self, players: list[Player]):
        pass

    def get_max_pages(self):
        if self.specified_guild is not None and (player := self.cog.pylav.player_manager.get(self.specified_guild)):
            players = [player]
        else:
            players = self.cog.pylav.player_manager.connected_players
        pages, left_over = divmod(len(players), self.per_page)
        if left_over:
            pages += 1
        return pages or 1

    def get_starting_index_and_page_number(self, menu: BaseMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: BaseMenu, player: Player) -> discord.Embed:
        idx_start, page_num = self.get_starting_index_and_page_number(menu)
        connect_dur = discord.utils.format_dt(player.connected_at, style="R")
        self.current_player = player
        guild_name = player.guild.name
        queue_len = player.queue.size()
        history_queue_len = player.history.size()
        server_owner = f"{player.guild.owner} ({player.guild.owner.id})"
        current_track = (
            await player.current.get_track_display_name(max_length=50, with_url=True)
            if player.current
            else _("Nothing playing")
        )

        listener_count = await asyncstdlib.sum(
            True async for m in asyncstdlib.iter(rgetattr(player, "channel.members", [])) if not m.bot
        )
        listeners = humanize_number(listener_count)
        current_track += "\n"

        field_values = "\n".join(
            f"**{i[0]}**: {i[1]}"
            for i in [
                (_("Server Owner"), server_owner),
                (_("Connected"), connect_dur),
                (_("Users in VC"), listeners),
                (
                    _("Queue Length"),
                    "{queue_length} {track_translation}".format(
                        queue_length=queue_len, track_translation=_("track" if queue_len == 1 else "tracks")
                    ),
                ),
                (
                    _("Queue History Length"),
                    "{count} {track_translation}".format(
                        count=history_queue_len,
                        track_translation=_("track") if history_queue_len == 1 else _("tracks"),
                    ),
                ),
            ]
        )

        current_track += field_values

        embed = await self.cog.pylav.construct_embed(messageable=menu.ctx, title=guild_name, description=current_track)

        embed.set_footer(
            text=_("Page {page_num}/{total_pages} | Playing in {playing} {server_translation}").format(
                page_num=humanize_number(page_num + 1),
                total_pages=humanize_number(self.get_max_pages()),
                playing=humanize_number(len(self.cog.pylav.player_manager.playing_players)),
                server_translation=_("server") if history_queue_len == 1 else _("servers"),
            )
        )
        return embed



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pylav',
 'pylav._internals',
 'pylav.constants',
 'pylav.constants.config',
 'pylav.core',
 'pylav.enums',
 'pylav.enums.plugins',
 'pylav.events',
 'pylav.events.plugins',
 'pylav.events.track',
 'pylav.exceptions',
 'pylav.extension',
 'pylav.extension.bundled_node',
 'pylav.extension.flowery',
 'pylav.extension.flowery.lyrics',
 'pylav.extension.m3u',
 'pylav.extension.radio',
 'pylav.extension.red',
 'pylav.extension.red.converters',
 'pylav.extension.red.ui',
 'pylav.extension.red.ui.buttons',
 'pylav.extension.red.ui.menus',
 'pylav.extension.red.ui.modals',
 'pylav.extension.red.ui.prompts',
 'pylav.extension.red.ui.selectors',
 'pylav.extension.red.ui.selectors.options',
 'pylav.extension.red.ui.sources',
 'pylav.extension.red.utils',
 'pylav.helpers',
 'pylav.helpers.discord',
 'pylav.helpers.discord.converters',
 'pylav.helpers.emojis',
 'pylav.helpers.format',
 'pylav.logging',
 'pylav.nodes',
 'pylav.nodes.api',
 'pylav.nodes.api.responses',
 'pylav.nodes.api.responses.filters',
 'pylav.nodes.api.responses.filters.plugins',
 'pylav.nodes.api.responses.plugins',
 'pylav.nodes.websockets',
 'pylav.players',
 'pylav.players.filters',
 'pylav.players.filters.plugins',
 'pylav.players.query',
 'pylav.players.query.obj',
 'pylav.players.tracks',
 'pylav.players.tracks.obj',
 'pylav.storage',
 'pylav.storage.controllers',
 'pylav.storage.controllers.players',
 'pylav.storage.database',
 'pylav.storage.database.cache',
 'pylav.storage.database.tables',
 'pylav.storage.migrations',
 'pylav.storage.migrations.high_level',
 'pylav.storage.migrations.high_level.always',
 'pylav.storage.migrations.high_level.one_offs',
 'pylav.storage.migrations.low_level',
 'pylav.storage.models',
 'pylav.storage.models.node',
 'pylav.storage.models.player',
 'pylav.type_hints',
 'pylav.utils',
 'pylav.utils.vendor',
 'pylav.utils.vendor.lavalink_py',
 'pylav.utils.vendor.redbot',
 'pylav.vendor']

package_data = \
{'': ['*'],
 'pylav': ['locales/*'],
 'pylav._internals': ['locales/*'],
 'pylav.constants': ['locales/*'],
 'pylav.constants.config': ['locales/*'],
 'pylav.core': ['locales/*'],
 'pylav.enums': ['locales/*'],
 'pylav.enums.plugins': ['locales/*'],
 'pylav.events': ['locales/*'],
 'pylav.events.plugins': ['locales/*'],
 'pylav.events.track': ['locales/*'],
 'pylav.exceptions': ['locales/*'],
 'pylav.extension': ['locales/*'],
 'pylav.extension.bundled_node': ['locales/*'],
 'pylav.extension.flowery': ['locales/*'],
 'pylav.extension.flowery.lyrics': ['locales/*'],
 'pylav.extension.m3u': ['locales/*'],
 'pylav.extension.radio': ['locales/*'],
 'pylav.extension.red': ['locales/*'],
 'pylav.extension.red.converters': ['locales/*'],
 'pylav.extension.red.ui': ['locales/*'],
 'pylav.extension.red.ui.buttons': ['locales/*'],
 'pylav.extension.red.ui.menus': ['locales/*'],
 'pylav.extension.red.ui.modals': ['locales/*'],
 'pylav.extension.red.ui.prompts': ['locales/*'],
 'pylav.extension.red.ui.selectors': ['locales/*'],
 'pylav.extension.red.ui.selectors.options': ['locales/*'],
 'pylav.extension.red.ui.sources': ['locales/*'],
 'pylav.extension.red.utils': ['locales/*'],
 'pylav.helpers': ['locales/*'],
 'pylav.helpers.discord': ['locales/*'],
 'pylav.helpers.discord.converters': ['locales/*'],
 'pylav.helpers.emojis': ['locales/*'],
 'pylav.helpers.format': ['locales/*'],
 'pylav.logging': ['locales/*'],
 'pylav.nodes': ['locales/*'],
 'pylav.nodes.api': ['locales/*'],
 'pylav.nodes.api.responses': ['locales/*'],
 'pylav.nodes.api.responses.filters': ['locales/*'],
 'pylav.nodes.api.responses.filters.plugins': ['locales/*'],
 'pylav.nodes.api.responses.plugins': ['locales/*'],
 'pylav.nodes.websockets': ['locales/*'],
 'pylav.players': ['locales/*'],
 'pylav.players.filters': ['locales/*'],
 'pylav.players.filters.plugins': ['locales/*'],
 'pylav.players.query': ['locales/*'],
 'pylav.players.tracks': ['locales/*'],
 'pylav.storage': ['locales/*'],
 'pylav.storage.controllers': ['locales/*'],
 'pylav.storage.controllers.players': ['locales/*'],
 'pylav.storage.database': ['locales/*'],
 'pylav.storage.database.cache': ['locales/*'],
 'pylav.storage.database.tables': ['locales/*'],
 'pylav.storage.migrations': ['locales/*'],
 'pylav.storage.migrations.high_level': ['locales/*'],
 'pylav.storage.migrations.high_level.always': ['locales/*'],
 'pylav.storage.migrations.high_level.one_offs': ['locales/*'],
 'pylav.storage.migrations.low_level': ['locales/*'],
 'pylav.storage.models': ['locales/*'],
 'pylav.storage.models.node': ['locales/*'],
 'pylav.storage.models.player': ['locales/*'],
 'pylav.type_hints': ['locales/*'],
 'pylav.utils': ['locales/*'],
 'pylav.utils.vendor': ['locales/*'],
 'pylav.utils.vendor.lavalink_py': ['locales/*'],
 'pylav.utils.vendor.redbot': ['locales/*'],
 'pylav.vendor': ['locales/*']}

install_requires = \
['APScheduler>=3.9.1.post1,<4.0.0',
 'Brotli>=1.0.9,<2.0.0',
 'PyYAML>=6.0,<7.0',
 'aiohttp-client-cache>=0.8.0,<0.9.0',
 'aiopath>=0.6.11,<0.7.0',
 'asyncpg>=0.27.0,<0.28.0',
 'asyncspotify>=0.12.1,<0.13.0',
 'asyncstdlib>=3.10.5,<4.0.0',
 'cached-property>=1.5.2,<2.0.0',
 'cashews[dill,redis]>=5.0.0,<6.0.0',
 'dacite>=1.7.0,<2.0.0',
 'deepdiff6>=6.2.0,<7.0.0',
 'expiringdict>=1.2.2,<2.0.0',
 'humanize>=4.4.0,<5.0.0',
 'iso8601>=1.1.0,<2.0.0',
 'netaddr>=0.8.0,<0.9.0',
 'numpy>=1.24.1,<2.0.0',
 'piccolo[all]>=0.103.0,<0.104.0',
 'pip>=22.3.1,<23.0.0',
 'setuptools>=65.6.3,<66.0.0',
 'tabulate[widechars]>=0.9.0,<0.10.0',
 'ujson>=5.6.0,<6.0.0',
 'wheel>=0.38.4,<0.39.0']

extras_require = \
{':extra == "redis" or extra == "all" or extra == "all-nored"': ['redis>=4.4.0,<5.0.0'],
 'all': ['orjson>=3.8.3,<4.0.0'],
 'all-nored': ['Red-Commons>=1.0.0,<2.0.0',
               'aiohttp[speedups]>=3.8.3,<4.0.0',
               'apsw>=3.40.0.0,<4.0.0.0',
               'importlib-metadata>=6.0.0,<7.0.0',
               'orjson>=3.8.3,<4.0.0',
               'orjson>=3.8.3,<4.0.0',
               'packaging>=22.0,<23.0',
               'platformdirs>=2.6.2,<3.0.0',
               'psutil>=5.9.4,<6.0.0',
               'python-dateutil>=2.8.2,<3.0.0',
               'pytz>=2022.7,<2023.0',
               'rich>=13.0.0,<14.0.0',
               'typing-extensions>=4.4.0,<5.0.0'],
 'all-nored:sys_platform != "win32" and platform_python_implementation == "CPython"': ['uvloop>=0.17.0,<0.18.0'],
 'nored': ['Red-Commons>=1.0.0,<2.0.0',
           'aiohttp[speedups]>=3.8.3,<4.0.0',
           'apsw>=3.40.0.0,<4.0.0.0',
           'importlib-metadata>=6.0.0,<7.0.0',
           'orjson>=3.8.3,<4.0.0',
           'packaging>=22.0,<23.0',
           'platformdirs>=2.6.2,<3.0.0',
           'psutil>=5.9.4,<6.0.0',
           'python-dateutil>=2.8.2,<3.0.0',
           'pytz>=2022.7,<2023.0',
           'rich>=13.0.0,<14.0.0',
           'typing-extensions>=4.4.0,<5.0.0'],
 'nored:sys_platform != "win32" and platform_python_implementation == "CPython"': ['uvloop>=0.17.0,<0.18.0'],
 'speed': ['orjson>=3.8.3,<4.0.0']}

setup_kwargs = {
    'name': 'py-lav',
    'version': '1.1.18',
    'description': ' A Lavalink WebSocket & API wrapper for discord.py',
    'long_description': "[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/Drapersniper/Py-Lav/master.svg)](https://results.pre-commit.ci/latest/github/Drapersniper/Py-Lav/master)\n[![GitHub license](https://img.shields.io/github/license/Drapersniper/Py-Lav.svg)](https://github.com/Drapersniper/Py-Lav/blob/master/LICENSE)\n[![Support Server](https://img.shields.io/discord/970987707834720266)](https://discord.com/invite/Sjh2TSCYQB)\n[![PyPi](https://img.shields.io/pypi/v/Py-Lav?style=plastic)](https://pypi.org/project/Py-Lav/)\n[![Documentation Status](https://readthedocs.org/projects/pylav/badge/?version=latest)](https://pylav.readthedocs.io/en/latest/?badge=latest)\n[![Crowdin](https://badges.crowdin.net/pylav/localized.svg)](https://crowdin.com/project/pylav)\n\n\n# Documentation\n### Installation\n - [Click Here](SETUP.md)\n---------------------------\n### Requirements\n- PostgresSQL 14 server\n  - MacOS: [PostgresSQL](https://www.postgresql.org/download/macosx/)\n  - Windows: [PostgresSQL](https://www.postgresql.org/download/windows/)\n  - Linux: [PostgresSQL](https://www.postgresql.org/download/linux/)\n- Python 3.11\n- [Discord.py](https://github.com/Rapptz/discord.py) 2.1.0+ bot\n- [Lavalink](https://github.com/freyacodes/Lavalink) v4.0.0+ server\n---------------------------\n## Supported sources\n### [Built-in](https://github.com/freyacodes/Lavalink):\n  - youtube\n  - soundcloud\n  - bandcamp\n  - twitch\n  - vimeo\n  - http\n  - local\n### With [LavaSrc](https://github.com/TopiSenpai/LavaSrc)\n  - spotify\n  - applemusic\n  - deezer\n### With [DuncteBot-plugin](https://github.com/DuncteBot/skybot-lavalink-plugin):\n  - getyarn.io\n  - clypit\n  - tts\n  - pornhub\n  - reddit\n  - ocremix\n  - tiktok\n  - mixcloud\n  - soundgasm\n### With [Google Cloud TTS](https://github.com/DuncteBot/tts-plugin):\n  - gcloud-tts\n### With [Sponsorblock-Plugin](https://github.com/Topis-Lavalink-Plugins/Sponsorblock-Plugin):\n  - sponsorblock\n\n## Supported Lavalink features\n  - Supports all features of [Lavalink](https://github.com/freyacodes/Lavalink)\n    - Filters\n    - IP Rotation\n    - Plug-ins\n\nFeatures\n---------------------------\n- Multiple node support\n  - Node region assignment based on IP\n- Track cache for the last 30 days to reduce the number of duplicated queries\n- Managed local node with auto-restart and auto update\n- Many helper methods and functions\n- Support for multiple cogs to access the library at once\n- Playlists and EQ saved at a library level to be shared with cogs\n- Player state persistence\n    - Upon library state being shutdown all player states are saved and restored on library startup\n- History of played tracks available for players\n- RadioBrowser.org API to retrieve radio stations available for players\n- Lyrics API to retrieve lyrics for tracks provided by [Flowery](https://flowery.pw)\n- M3U, PLS and PYLAV text file parser to convert contents into a playlist (p.s. Lavalink must support the format/coded of files still)\n\n\nCredits\n---------------------------\n- [Topi](https://github.com/TopiSenpai) for all the work done to Lavalink and implementing direct requests to make PyLav even better.\n- [Devoxin - Lavalink.py](https://github.com/Devoxin/Lavalink.py) for the original ideas for implementation.\n- [Ryan](https://github.com/ryan5453) for the amazing [Lyrics API](https://flowery.pw) used for lyrics.\n- [globocom/m3u8](https://github.com/globocom/m3u8) for the M3U8 parser which I made asynchronous found in [m3u8_parser](pylav/extension/m3u).\n- [andreztz/pyradios](https://github.com/andreztz/pyradios) for the radio parser which I made asynchronous found in [radio](pylav/extension/radio).\n- [Lifeismana](https://github.com/Lifeismana) for the custom Red-DiscordBot docker image which added Python3.11 support until Phasecore's image is updated.\n",
    'author': 'Draper',
    'author_email': '27962761+Drapersniper@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Drapersniper/PyLav',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import functools
import hashlib
from abc import ABC
from collections.abc import MutableMapping
from typing import Any

from redbot.core import commands

from pylav.type_hints.dict_typing import JSON_DICT_TYPE


def rsetattr(obj: object, attr: str, val: Any) -> None:
    pre, _, post = attr.rpartition(".")
    setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj: object, attr: str, *args: Any) -> Any:
    def _getattr(obj2, attr2):
        return getattr(obj2, attr2, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def recursive_merge(d1: JSON_DICT_TYPE, d2: JSON_DICT_TYPE) -> JSON_DICT_TYPE:
    """
    Update two dicts of dicts recursively,
    if either mapping has leaves that are non-dicts,
    the second's leaf overwrites the first's.
    """
    for k, v in d1.items():
        if k in d2 and all(isinstance(e, MutableMapping) for e in (v, d2[k])):
            d2[k] = recursive_merge(v, d2[k])
    return d1 | d2


class Mutator:
    def __init__(self, obj: Any):
        self.obj = obj
        self.name = obj.name if hasattr(obj, "name") else str(obj.__class__.__name__)
        if hasattr(obj, "id"):
            self.id = str(obj.id)
        else:
            i = hashlib.md5()
            i.update(self.name.encode())
            i.update(id(obj).encode())
            self.id = str(int(i.hexdigest(), 16))[:16]


class CompositeMetaClass(type(commands.Cog), type(ABC)):
    """
    This allows the metaclass used for proper type detection to
    coexist with discord.py's metaclass
    """



from __future__ import annotations

import asyncio
from collections.abc import Callable
from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.helpers import emojis
from pylav.helpers.format.strings import shorten_string
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class NavigateButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        emoji: str | discord.PartialEmoji,
        direction: int | Callable[[], int],
        row: int = None,
        label: str = None,
    ):

        super().__init__(style=style, emoji=emoji, row=row, label=label)
        self.cog = cog
        self._direction = direction

    @property
    def direction(self) -> int:
        return self._direction if isinstance(self._direction, int) else self._direction()

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        max_pages = self.view.source.get_max_pages()
        if self.direction == 0:
            self.view.current_page = 0
        elif self.direction == max_pages:
            self.view.current_page = max_pages - 1
        else:
            self.view.current_page += self.direction

        if self.view.current_page >= max_pages:
            self.view.current_page = 0
        elif self.view.current_page < 0:
            self.view.current_page = max_pages - 1

        kwargs = await self.view.get_page(self.view.current_page)
        await self.view.prepare()
        await self.view.message.edit(view=self.view, **kwargs)


class CloseButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.MINIMIZE,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            return await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        self.view.cancelled = True
        self.view.stop()
        await self.view.on_timeout()


class YesButton(discord.ui.Button):
    interaction: DISCORD_INTERACTION_TYPE

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(style=style, emoji=None, row=row, label=shorten_string(max_length=100, string=_("Yes")))
        self.responded = asyncio.Event()
        self.cog = cog
        self.interaction = None  # type: ignore

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
            return
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        self.responded.set()
        self.interaction = interaction


class NoButton(discord.ui.Button):
    interaction: DISCORD_INTERACTION_TYPE

    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(style=style, emoji=None, row=row, label=shorten_string(max_length=100, string=_("No")))
        self.responded = asyncio.Event()
        self.cog = cog
        self.interaction = None  # type: ignore

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
            return
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        self.responded.set()
        self.interaction = interaction


class DoneButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.CHECK,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            return await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        self.view.done = True
        self.view.cancelled = False
        self.view.stop()
        await self.view.on_timeout()


class LabelButton(discord.ui.Button):
    def __init__(
        self,
        disconnect_type_translation: str,
        multiple=True,
        row: int = None,
    ):
        super().__init__(
            style=discord.ButtonStyle.secondary,
            emoji=None,
            row=row,
        )
        self.label = _("Disconnect {} {}").format(
            disconnect_type_translation, _("players") if multiple else _("player")
        )


class RefreshButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji="\N{ANTICLOCKWISE DOWNWARDS AND UPWARDS OPEN CIRCLE ARROWS}",
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        await self.view.message.edit(view=self.view, **kwargs)



from __future__ import annotations

import hashlib
import struct
import typing
import uuid
from functools import total_ordering
from typing import Any, Self

import asyncstdlib
import discord
from cached_property import cached_property  # type: ignore
from cashews import Cache  # type: ignore

from pylav.constants.regex import SQUARE_BRACKETS, STREAM_TITLE
from pylav.exceptions.track import InvalidTrackException, TrackNotFoundException
from pylav.extension.flowery.lyrics import Error, Lyrics
from pylav.helpers.misc import MISSING
from pylav.nodes.api.responses.track import Track as APITrack
from pylav.players.query.obj import Query
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

if typing.TYPE_CHECKING:
    from pylav.core.client import Client
    from pylav.nodes.node import Node
    from pylav.players.player import Player


__CLIENT: Client | None = None


# noinspection SpellCheckingInspection
@total_ordering
class Track:
    __slots__ = (
        "_node",
        "_is_partial",
        "_query",
        "_extra",
        "_unique_id",
        "__clear_cache_task",
        "_skip_segments",
        "_requester_id",
        "_requester",
        "_updated_query",
        "_id",
        "_extra",
        "_encoded",
        "_raw_data",
        "_processed",
    )
    __CLIENT: Client | None = None

    def __init__(
        self,
        node: Node,
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> None:
        """This class should not be instantiated directly. Use :meth:`Track.build_track` instead."""
        self._encoded: str | None = None
        self._is_partial = False
        self._requester = requester
        self._node = node
        self._query = query
        self._extra = extra
        self._skip_segments = skip_segments or []
        self._unique_id = hashlib.md5()
        self._updated_query = None
        self._id = str(uuid.uuid4())
        self._raw_data: JSON_DICT_TYPE = {}
        self._processed = None

        self._process_init()

    @property
    def client(self) -> Client:
        """Get the client"""
        global __CLIENT
        return self.__CLIENT or __CLIENT

    @classmethod
    def attach_client(cls, client: Client) -> None:
        global __CLIENT
        __CLIENT = cls.__CLIENT = client

    def __int__(self) -> int:
        return 0

    def __gt__(self, other: Any) -> bool:
        return False

    def __lt__(self, other: Any) -> bool:
        return False

    def __ge__(self, other: Any) -> bool:
        return True

    def __le__(self, other: Any) -> bool:
        return True

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Track):
            return self._id == other._id
        raise NotImplemented

    def __ne__(self, other: Any) -> bool:
        x = self.__eq__(other)
        return not x

    def __hash__(self):
        return hash((self.unique_identifier,))

    def __getitem__(self, name: str) -> Any:
        return super().__getattribute__(name)

    def __repr__(self) -> str:
        return f"<Track identifier={self._id} encoded={self.encoded}>"

    def _process_init(self) -> None:
        if self._requester is None:
            self._requester_id = self.client.bot.user.id
        elif isinstance(self._requester, int):
            self._requester_id = self._requester
        else:
            self._requester_id = self._requester.id

        if "partial" in self._extra:
            self._is_partial = self._extra["partial"]

        if self._query is not None:
            self.timestamp = self.timestamp or self._query.start_time

        self._extra.pop("track", None)

    @classmethod
    async def build_track(
        cls,
        node: Node,
        data: Track | APITrack | dict[str, Any] | str | None,
        query: Query | None,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        lazy: bool = False,
        **extra: Any,
    ) -> Track:
        """Builds a track object from the given data.

        Parameters
        ----------
        node: None
            The node that the track is being built for.
        data: :class:`Track` | :class:`LavalinkTrackObject` | :class:`dict` | :class:`str` | None
            The data to build the track from.
        query: :class:`Query`
            The query that the track was built from.
        skip_segments: :class:`list`[:class:`str`]
            The segments to skip when playing the track.
        requester: :class:`discord.abc.User` | :class:`int`
            The user that requested the track.
        lazy: :class:`bool`
            Whether to build decode via the bot or Lavalink
        **extra: Any
            Extra data to add to the track.
        """

        # Check if data is a Track object and return it if it is
        if isinstance(data, cls):
            return data

        # Check if data is a LavalinkTrackObject and process it.
        if isinstance(data, APITrack):
            return cls._from_lavalink_track_object(node, data, query, skip_segments, requester, **extra)

        if isinstance(data, dict):
            try:
                return cls._from_mapping(node, data, query, skip_segments, requester, **extra)
            except KeyError as ke:
                (missing_key,) = ke.args
                raise InvalidTrackException(
                    f"Cannot build a track from partial data! (Missing key: {missing_key})"
                ) from ke

        if data is None or (isinstance(data, str) and data == MISSING):
            return cls._from_query(node, query, skip_segments, requester, **extra)

        if isinstance(data, str):
            return await cls._from_base64_string(node, data, skip_segments, requester, lazy=lazy, **extra)

        raise TypeError(f"Expected Track, LavalinkTrackObject, dict, or str, got {type(data).__name__}")

    @classmethod
    def _from_lavalink_track_object(
        cls,
        node: Node,
        data: APITrack,
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> Track:
        instance = cls(node, query, skip_segments, requester, **extra)
        instance._encoded = data.encoded
        instance._extra = extra
        instance._raw_data = extra.get("raw_data", {})
        instance._unique_id = hashlib.md5()
        instance._unique_id.update(instance._encoded.encode())
        instance._is_partial = False
        instance._processed = data
        return instance

    @classmethod
    def _from_mapping(
        cls,
        node: Node,
        data: dict[str, Any],
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> Track:
        instance = cls(node, query, skip_segments, requester, **extra)
        instance._encoded = data["encoded"]
        instance._is_partial = not bool(data["encoded"])
        instance._raw_data = data.get("raw_data", {}) or extra.get("raw_data", {})
        if instance._encoded is None:
            raise InvalidTrackException("Cannot build a track from partial data! (Missing key: encoded)")
        instance._unique_id = hashlib.md5()
        instance._unique_id.update(instance._encoded.encode())
        return instance

    @classmethod
    async def _from_base64_string(
        cls,
        node: Node,
        data: str,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        lazy: bool = False,
        **extra: Any,
    ) -> Track:
        track_obj = await cls.__CLIENT.decode_track(data, raise_on_failure=True, lazy=lazy)
        query_obj = await Query.from_string(track_obj.info.uri)
        return cls._from_lavalink_track_object(
            node=node, data=track_obj, query=query_obj, skip_segments=skip_segments, requester=requester, **extra
        )

    @classmethod
    def _from_query(
        cls,
        node: Node,
        query: Query,
        skip_segments: list[str] | None = None,
        requester: discord.abc.User | int | None = None,
        **extra: Any,
    ) -> Track:
        instance = cls(node, query, skip_segments, requester, **extra)
        instance._extra = extra
        instance._raw_data = extra.get("raw_data", {})
        instance._is_partial = True
        return instance

    def _copy_with_extras(self, node: Node, **extra: Any) -> Track:
        instance = self.__class__(node, self._query, None, None)
        instance._extra = {**self._extra, **extra}
        instance._encoded = self._encoded
        instance._raw_data = self._raw_data
        instance._unique_id = self._unique_id
        instance._is_partial = self._is_partial
        instance._processed = self._processed
        return instance

    @property
    def skip_segments(self) -> list[str]:
        """The segments to skip when playing the track."""
        return self._skip_segments

    @property
    def encoded(self) -> str | None:
        return self._encoded

    @property
    def timestamp(self) -> int:
        return self._extra.get("timestamp", 0)

    @timestamp.setter
    def timestamp(self, value: int) -> None:
        self._extra["timestamp"] = value

    @property
    def requester_id(self) -> int:
        return self._requester_id

    @property
    def requester(self) -> discord.User | None:
        return self.client.bot.get_user(self.requester_id)

    @property
    def last_known_position(self) -> int:
        return self._extra.get("last_known_position", 0)

    @last_known_position.setter
    def last_known_position(self, value: int) -> None:
        self._extra["last_known_position"] = value

    @property
    def id(self) -> str:
        return self._id

    @property
    def unique_identifier(self) -> str:
        return self._unique_id.hexdigest()

    @property
    def is_partial(self) -> bool:
        return bool(self._is_partial and not self.encoded)

    async def identifier(self) -> str | None:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.identifier

    async def is_seekable(self) -> bool:
        return False if self.is_partial else (await self.fetch_full_track_data()).info.isSeekable

    async def duration(self) -> int:
        return 0 if self.is_partial else (await self.fetch_full_track_data()).info.length

    length = duration

    async def stream(self) -> bool:
        return False if self.is_partial else (await self.fetch_full_track_data()).info.isStream

    async def title(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.title

    async def uri(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.uri

    async def author(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.author

    async def source(self) -> str:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.sourceName

    async def artworkUrl(self) -> str | None:  # noqa:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.artworkUrl

    async def isrc(self) -> str | None:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).info.isrc

    async def probe_info(self) -> str | None:
        return MISSING if self.is_partial else (await self.fetch_full_track_data()).pluginInfo.probeInfo

    async def query(self) -> Query:
        if self.encoded and self._updated_query is None:
            assert self.encoded is not None
            self._updated_query = await Query.from_base64(self.encoded, lazy=True)
            self._updated_query.merge(
                query=self._query,
                start_time=True,
                index=True,
                source=True,
                recursive=True,
                search=True,
            )
            self._query = self._updated_query
            self.timestamp = self.timestamp or self._query.start_time
        return self._query

    async def is_clypit(self) -> bool:
        return (await self.query()).is_clypit

    async def is_getyarn(self) -> bool:
        return (await self.query()).is_getyarn

    async def is_mixcloud(self) -> bool:
        return (await self.query()).is_mixcloud

    async def is_ocremix(self) -> bool:
        return (await self.query()).is_ocremix

    async def is_pornhub(self) -> bool:
        return (await self.query()).is_pornhub

    async def is_reddit(self) -> bool:
        return (await self.query()).is_reddit

    async def is_soundgasm(self) -> bool:
        return (await self.query()).is_soundgasm

    async def is_tiktok(self) -> bool:
        return (await self.query()).is_tiktok

    async def is_spotify(self) -> bool:
        return (await self.query()).is_spotify

    async def is_apple_music(self) -> bool:
        return (await self.query()).is_apple_music

    async def is_bandcamp(self) -> bool:
        return (await self.query()).is_bandcamp

    async def is_youtube(self) -> bool:
        return (await self.query()).is_youtube

    async def is_youtube_music(self) -> bool:
        return (await self.query()).is_youtube_music

    async def is_soundcloud(self) -> bool:
        return (await self.query()).is_soundcloud

    async def is_twitch(self) -> bool:
        return (await self.query()).is_twitch

    async def is_http(self) -> bool:
        return (await self.query()).is_http

    async def is_local(self) -> bool:
        return (await self.query()).is_local

    async def is_niconico(self) -> bool:
        return (await self.query()).is_niconico

    async def is_vimeo(self) -> bool:
        return (await self.query()).is_vimeo

    async def is_search(self) -> bool:
        return (await self.query()).is_search

    async def is_album(self) -> bool:
        return (await self.query()).is_album

    async def is_playlist(self) -> bool:
        return (await self.query()).is_playlist

    async def is_single(self) -> bool:
        return (await self.query()).is_single

    async def is_speak(self) -> bool:
        return (await self.query()).is_speak

    async def is_gctts(self) -> bool:
        return (await self.query()).is_gctts

    async def is_deezer(self) -> bool:
        return (await self.query()).is_deezer

    async def requires_capability(self) -> str:
        return q.requires_capability if (q := await self.query()) else "youtube"

    async def query_identifier(self) -> str:
        return (await self.query()).query_identifier

    async def query_source(self) -> str:
        return (await self.query()).source

    async def mix_playlist_url(self) -> str | None:
        if not await self.identifier():
            return None
        if await self.source() == "youtube":
            return f"https://www.youtube.com/watch?v={self.identifier}&list=RD{self.identifier}"
        return None

    async def fetch_full_track_data(
        self,
    ) -> APITrack:
        if self._processed:
            return self._processed
        if self.encoded:
            self._processed = await self.client.decode_track(self.encoded)
        else:
            await self.search()
        return self._processed

    async def to_dict(self) -> dict[str, Any]:
        """
        Returns a dict representation of this Track.
        Returns
        -------
        :class:`dict`
            The dict representation of this Track.
        """
        return {
            "encoded": self.encoded,
            "query": await self.query_identifier(),
            "requester": self.requester.id if self.requester else self.requester_id,
            "skip_segments": self._skip_segments,
            "extra": {
                "timestamp": self.timestamp,
                "last_known_position": self.last_known_position,
                "partial": self._is_partial,
            },
            "raw_data": self._raw_data,
        }

    async def search(self, bypass_cache: bool = False) -> Self:
        _query = await Query.from_string(self._query)
        _query.merge(
            query=self._query,
            start_time=True,
            index=True,
            source=True,
            recursive=True,
            search=True,
        )
        self._query = _query
        self.timestamp = self.timestamp or self._query.start_time
        response = await self.client._get_tracks(await self.query(), first=True, bypass_cache=bypass_cache)
        if not response or not response.tracks:
            raise TrackNotFoundException(f"No tracks found for query {await self.query_identifier()}")
        track = response.tracks[0]
        self._encoded = track.encoded
        assert isinstance(self._encoded, str)
        self._unique_id = hashlib.md5()
        self._unique_id.update(self.encoded.encode())
        self._is_partial = False
        self._processed = track

    async def search_all(self, player: Player, requester: int, bypass_cache: bool = False) -> list[Track]:
        _query = await Query.from_string(self._query)
        _query.merge(
            query=self._query,
            start_time=True,
            index=True,
            source=True,
            recursive=True,
            search=True,
        )
        self._query = _query
        response = await player.node.get_track(
            await self.query(), bypass_cache=bypass_cache, first=self._query.is_search
        )
        if not response or not response.tracks:
            raise TrackNotFoundException(f"No tracks found for query {await self.query_identifier()}")
        return [
            await Track.build_track(data=track, node=player.node, query=self._query, requester=requester)
            for track in response.tracks
        ]

    async def _icyparser(self, url: str) -> str | None:
        try:
            async with self._node.session.get(url, headers={"Icy-MetaData": "1"}) as resp:
                metaint = int(resp.headers["icy-metaint"])
                async for __ in asyncstdlib.iter(range(5)):
                    await resp.content.readexactly(metaint)
                    metadata_length = struct.unpack("B", await resp.content.readexactly(1))[0] * 16
                    metadata = await resp.content.readexactly(metadata_length)
                    if not (m := STREAM_TITLE.search(metadata.rstrip(b"\0"))):
                        return None
                    if title := m.group(1):
                        title = title.decode("utf-8", errors="replace")
                        return title
        except Exception:  # noqa
            return None
        return None

    async def get_track_display_name(
        self,
        max_length: int | None = None,
        author: bool = True,
        unformatted: bool = False,
        with_url: bool = False,
        escape: bool = True,
    ) -> str:
        if unformatted:
            return await self.get_track_display_name_unformatted(max_length=max_length, author=author, escape=escape)
        else:
            return await self.get_track_display_name_formatted(
                max_length=max_length, author=author, with_url=with_url, escape=escape
            )

    async def get_track_display_name_unformatted(
        self,
        max_length: int | None = None,
        author: bool = True,
        escape: bool = True,
    ) -> str:
        if self.is_partial:
            track_name = await self.get_partial_track_display_name(max_length=max_length)
        else:
            track_name = await self.get_full_track_display_name(max_length=max_length, author=author)
        return self._maybe_escape_markdown(text=track_name, escape=escape)

    async def get_track_display_name_formatted(
        self,
        max_length: int | None = None,
        author: bool = True,
        with_url: bool = False,
        escape: bool = True,
    ) -> str:

        if self.is_partial:
            track_name = await self.get_partial_track_display_name(
                max_length=(max_length - 8) if with_url and isinstance(max_length, int) else max_length
            )
            track_name = self._maybe_escape_markdown(text=track_name, escape=escape)
            if with_url and (query := await self.query()):
                if not query.is_single and not query.is_custom_playlist:
                    track_name = f"**[{track_name}]({query.query_identifier})**"
        else:
            track_name = await self.get_full_track_display_name(
                max_length=(max_length - 8) if with_url and isinstance(max_length, int) else max_length, author=author
            )
            track_name = self._maybe_escape_markdown(text=track_name, escape=escape)
            if with_url:
                track_name = f"**[{track_name}]({await self.uri()})**"

        return track_name

    async def get_partial_track_display_name(self, max_length: int | None = None) -> str:
        query = await self.query()
        track_name = await query.query_to_queue(max_length, partial=True)
        track_name = SQUARE_BRACKETS.sub("", track_name).strip()
        if max_length is not None and len(track_name) > (max_length - 1):
            max_length -= 1
            return f"{track_name[:max_length]}\N{HORIZONTAL ELLIPSIS}"
        return track_name

    async def get_full_track_display_name(self, max_length: int | None = None, author: bool = True) -> str:
        author_string = f" - {await self.author()}" if author else ""
        return (
            await self.get_local_query_track_display_name(
                max_length=max_length,
                author_string=author_string,
                unknown_author=await self.author() != "Unknown artist",
                unknown_title=await self.title() != "Unknown title",
            )
            if await self.query() and await self.is_local()
            else await self.get_external_query_track_display_name(max_length=max_length, author_string=author_string)
        )

    async def get_local_query_track_display_name(
        self,
        author_string: str,
        unknown_author: bool,
        unknown_title: bool,
        max_length: int | None = None,
    ) -> str:
        if not (unknown_title and unknown_author):
            track_name = f"{await self.title()}{author_string}"
            track_name = SQUARE_BRACKETS.sub("", track_name).strip()
            if max_length and len(track_name) > max_length:
                track_name = f"{track_name[:max_length]}\N{HORIZONTAL ELLIPSIS}"
            elif not max_length:
                track_name += f"\n{await (await self.query()).query_to_string(add_ellipsis=False)} "
        else:
            track_name = await (await self.query()).query_to_string(max_length, name_only=True)
            track_name = SQUARE_BRACKETS.sub("", track_name).strip()
        return track_name

    async def get_external_query_track_display_name(self, author_string: str, max_length: int | None = None) -> str:
        title = await self.title()

        if await self.stream():
            icy = await self._icyparser(await self.uri())
            track_name = icy or f"{title}{author_string}"
        elif (await self.author()).lower() not in title.lower():
            track_name = f"{title}{author_string}"
        else:
            track_name = title

        track_name = SQUARE_BRACKETS.sub("", track_name).strip()
        if max_length is not None and len(track_name) > (max_length - 1):
            max_length -= 1
            return f"{track_name[:max_length]}\N{HORIZONTAL ELLIPSIS}"
        return track_name

    @staticmethod
    def _maybe_escape_markdown(text: str, escape: bool = True) -> str:
        return discord.utils.escape_markdown(text) if escape else text

    async def fetch_lyrics(self) -> tuple[bool, Lyrics | Error | None]:
        if isrc := await self.isrc():
            return True, await self.client.flowery_api.lyrics.get_lyrics(isrc=isrc)
        elif not self.is_partial:
            if await self.is_spotify():
                return True, await self.client.flowery_api.lyrics.get_lyrics(spotify_id=await self.identifier())
            elif await self.source() in {"deezer", "applemusic"}:
                return True, await self.client.flowery_api.lyrics.get_lyrics(
                    query=f"{await self.title()} artist:{await self.author()}"
                )
            else:
                return False, await self.client.flowery_api.lyrics.get_lyrics(query=await self.title())
        return False, None
